import numpy as _np
from netket import random as _random

from .abstract_sampler import AbstractSampler
from .metropolis_hastings import *
from .._C_netket import sampler as c_sampler

from numba import jit, int64, float64
from .._jitclass import jitclass


@jitclass([("local_states", float64[:]), ("size", int64), ("n_states", int64)])
class _local_kernel:
    def __init__(self, local_states, size):
        self.local_states = _np.sort(
            _np.asarray(local_states, dtype=_np.float64))
        self.size = size
        self.n_states = self.local_states.size

    def apply(self, state, state_1, log_prob_corr):

        for i in range(state.shape[0]):
            state_1[i] = state[i]

            si = _random.randint(0, self.size)

            rs = _random.randint(0, self.n_states - 1)

            state_1[i, si] = self.local_states[
                rs + (self.local_states[rs] >= state[i, si])
            ]

        log_prob_corr[:] = 0.0


class MetropolisLocal(AbstractSampler):
    """
    Sampler acting on one local degree of freedom.

    This sampler acts locally only on one local degree of freedom :math:`s_i`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s_N`,
    where :math:`s^\prime_i \\neq s_i`.

    The transition probability associated to this
    sampler can be decomposed into two steps:

    1. One of the site indices :math:`i = 1\dots N` is chosen
    with uniform probability.
    2. Among all the possible (:math:`m`) values that :math:`s_i` can take,
    one of them is chosen with uniform probability.

    For example, in the case of spin :math:`1/2` particles, :math:`m=2`
    and the possible local values are :math:`s_i = -1,+1`.
    In this case then :class:`MetropolisLocal` is equivalent to flipping a random spin.

    In the case of bosons, with occupation numbers
    :math:`s_i = 0, 1, \dots n_{\mathrm{max}}`, :class:`MetropolisLocal`
    would pick a random local occupation number uniformly between :math:`0`
    and :math:`n_{\mathrm{max}}`.
    """

    def __init__(
        self, machine, n_chains=16, sweep_size=None, batch_size=None, backend=None
    ):
        """

         Constructs a new :class:`MetropolisLocal` sampler given a machine.

         Args:
            machine: A machine :math:`\Psi(s)` used for the sampling.
                     The probability distribution being sampled
                     from is :math:`F(\Psi(s))`, where the function
                     :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.

            n_chains:   The number of Markov Chain to be run in parallel on a single process.
            sweep_size: The number of exchanges that compose a single sweep.
                        If None, sweep_size is equal to the number of degrees of freedom (n_visible).
            batch_size: The batch size to be used when calling log_val on the given Machine.
                        If None, batch_size is equal to the number Markov chains (n_chains).


         Examples:
             Sampling from a RBM machine in a 1D lattice of spin 1/2

             >>> import netket as nk
             >>>
             >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
             >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
             >>>
             >>> # RBM Spin Machine
             >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
             >>>
             >>> # Construct a MetropolisLocal Sampler
             >>> sa = nk.sampler.MetropolisLocal(machine=ma)
             >>> print(sa.machine.hilbert.size)
             100
        """
        if "_C_netket.machine" in str(type(machine)) and backend != "py":
            self.sampler = c_sampler.MetropolisLocal(
                machine=machine,
                n_chains=n_chains,
                sweep_size=sweep_size,
                batch_size=batch_size,
            )
        else:
            self.sampler = PyMetropolisHastings(
                machine,
                _local_kernel(
                    _np.asarray(
                        machine.hilbert.local_states), machine.hilbert.size
                ),
                n_chains,
                sweep_size,
                batch_size,
            )
        super().__init__(machine, n_chains)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance


class MetropolisLocalPt(AbstractSampler):
    """
    This sampler performs parallel-tempering
    moves in addition to the local moves implemented in `MetropolisLocal`.
    The number of replicas can be chosen by the user.
    """

    def __init__(self, machine, n_replicas=16, sweep_size=None, batch_size=None):
        """
        Args:
             machine: A machine :math:`\Psi(s)` used for the sampling.
                      The probability distribution being sampled
                      from is :math:`F(\Psi(s))`, where the function
                      :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
             n_replicas: The number of replicas used for parallel tempering.
             sweep_size: The number of exchanges that compose a single sweep.
                         If None, sweep_size is equal to the number of degrees of freedom (n_visible).

        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.MetropolisLocalPt(
                machine=machine, n_replicas=n_replicas, sweep_size=sweep_size
            )
        else:
            raise ValueError(
                """Parallel Tempering samplers are not yet implemented
                for pure python machines"""
            )
        super().__init__(machine, 1)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance



import numpy as _np
from .abstract_sampler import AbstractSampler

from .._C_netket import sampler as c_sampler
from .._C_netket.utils import random_engine
from ..stats import mean as _mean

from netket import random as _random

from numba import jit, int64, float64
from .._jitclass import jitclass


class PyMetropolisHastings(AbstractSampler):
    """
    ``MetropolisHastings`` is a generic Metropolis-Hastings sampler using
    a local transition kernel to perform moves in the Markov Chain.
    The transition kernel is used to generate
    a proposed state :math:`s^\prime`, starting from the current state :math:`s`.
    The move is accepted with probability

    .. math::
    A(s\rightarrow s^\prime) = \mathrm{min}\left (1,\frac{P(s^\prime)}{P(s)} F(e^{L(s,s^\prime)})\right),

    where the probability being sampled is :math:`F(\Psi(s))` (by default :math:`F(x)=|x|^2`)
    and :math:`L(s,s^\prime)` is a correcting factor computed by the transition kernel.
    """

    def __init__(
        self, machine, transition_kernel, n_chains=16, sweep_size=None, batch_size=None
    ):
        """
        Constructs a new ``MetropolisHastings`` sampler given a machine and
        a transition kernel.

        Args:
            machine: A machine :math:`\Psi(s)` used for the sampling.
                          The probability distribution being sampled
                          from is :math:`F(\Psi(s))`, where the function
                          $$F(X)$$, is arbitrary, by default :math:`F(X)=|X|^2`.
            transition_kernel: A function to generate a transition.
                          This should take as an input the current state (in batches)
                          and return a modified state (also in batches).
                          This function must also return an array containing the
                          `log_prob_corrections` :math:`L(s,s^\prime)`.
            n_chains: The number of Markov Chain to be run in parallel on a single process.
            sweep_size: The number of exchanges that compose a single sweep.
                        If None, sweep_size is equal to the number of degrees of freedom (n_visible).
            batch_size: The batch size to be used when calling log_val on the given Machine.
                        If None, batch_size is equal to the number Markov chains (n_chains).

        """

        self.machine = machine
        self.n_chains = n_chains

        self.sweep_size = sweep_size

        self._kernel = transition_kernel

        self.machine_pow = 2.0

        super().__init__(machine, n_chains)

    @property
    def n_chains(self):
        return self._n_chains

    @n_chains.setter
    def n_chains(self, n_chains):
        if n_chains < 0:
            raise ValueError("Expected a positive integer for n_chains ")

        self._n_chains = n_chains

        self._state = _np.zeros((n_chains, self._n_visible))
        self._state1 = _np.copy(self._state)

        self._log_values = _np.zeros(n_chains, dtype=_np.complex128)
        self._log_values_1 = _np.zeros(n_chains, dtype=_np.complex128)
        self._log_prob_corr = _np.zeros(n_chains)

    @property
    def machine_pow(self):
        return self._machine_pow

    @machine_pow.setter
    def machine_pow(self, m_power):
        self._machine_pow = m_power

    @property
    def sweep_size(self):
        return self._sweep_size

    @sweep_size.setter
    def sweep_size(self, sweep_size):
        self._sweep_size = sweep_size if sweep_size != None else self._n_visible
        if self._sweep_size < 0:
            raise ValueError("Expected a positive integer for sweep_size ")

    @property
    def machine(self):
        return self._machine

    @machine.setter
    def machine(self, machine):
        self._machine = machine
        self._n_visible = machine.hilbert.size
        self._hilbert = machine.hilbert

    def reset(self, init_random=False):
        if init_random:
            for state in self._state:
                self._hilbert.random_vals(state, random_engine())
        self.machine.log_val(self._state, out=self._log_values)

        self._accepted_samples = 0
        self._total_samples = 0

    @staticmethod
    @jit(nopython=True)
    def acceptance_kernel(
        state, state1, log_values, log_values_1, log_prob_corr, machine_pow
    ):
        accepted = 0

        for i in range(state.shape[0]):
            prob = _np.exp(
                machine_pow *
                (log_values_1[i] - log_values[i] + log_prob_corr[i]).real
            )

            if prob > _random.uniform(0, 1):
                log_values[i] = log_values_1[i]
                state[i] = state1[i]
                accepted += 1

        return accepted

    def __next__(self):

        _log_val = self.machine.log_val
        _acc_kernel = self.acceptance_kernel
        _state = self._state
        _state1 = self._state1
        _log_values = self._log_values
        _log_values_1 = self._log_values_1
        _log_prob_corr = self._log_prob_corr
        _machine_pow = self._machine_pow
        _accepted_samples = self._accepted_samples
        _t_kernel = self._kernel.apply

        for sweep in range(self.sweep_size):

            # Propose a new state using the transition kernel
            _t_kernel(_state, _state1, _log_prob_corr)

            _log_val(_state1, out=_log_values_1)

            # Acceptance Kernel
            acc = _acc_kernel(
                _state,
                _state1,
                _log_values,
                _log_values_1,
                _log_prob_corr,
                _machine_pow,
            )

            _accepted_samples += acc

        self._total_samples += self.sweep_size * self.n_chains

        return self._state

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return _mean(self._accepted_samples) / _mean(self._total_samples)



import numpy as _np

import netket as _nk
from ._C_netket import MPI as _MPI
from .operator import local_values as _local_values
from .operator import der_local_values as _der_local_values
from netket.stats import (
    statistics as _statistics,
    covariance_sv as _covariance_sv,
    mean as _mean,
    subtract_mean as _subtract_mean,
)

from netket.vmc_common import info
from netket.abstract_variational_driver import AbstractVariationalDriver


class SteadyState(AbstractVariationalDriver):
    """
    Variational steady-state search by minimization of the L2,2 norm of L\rho using
    Variational Monte Carlo (VMC).
    """

    def __init__(
        self,
        lindblad,
        sampler,
        optimizer,
        n_samples,
        n_discard=None,
        sr=None,
        sampler_obs=None,
        n_samples_obs=None,
        n_discard_obs=None,
    ):
        """
        Initializes the driver class.

        Args:
            liouvillian: The liouvillian of the system.
            sampler: The Monte Carlo sampler for the density matrix.
            optimizer: Determines how optimization steps are performed given the
                bare energy gradient. This parameter supports three different kinds of inputs,
                which are described in the docs of `make_optimizer_fn`.
            n_samples: Number of Markov Chain Monte Carlo sweeps to be
                performed at each step of the optimization.
            n_discard (int, optional): Number of sweeps to be discarded at the
                beginning of the sampling, at each step of the optimization.
                Defaults to 10% of the number of samples allocated to each MPI node.
            sr (SR, optional): Determines whether and how stochastic reconfiguration
                is applied to the bare energy gradient before performing applying
                the optimizer. If this parameter is not passed or None, SR is not used.
            sampler_obs: The Monte Carlo sampler for the diagonal of the density matrix, used
                to compute observables (default: same as sampler).
            n_samples_obs: n_samples for the observables (default: n_samples)
            n_discard_obs: n_discard for the observables (default: n_discard)

        """
        super(SteadyState, self).__init__(
            sampler.machine, optimizer, minimized_quantity_name="LdagL"
        )  #'\u3008L\u2020L\u3009')

        self._lind = lindblad
        self._machine_obs = sampler_obs.machine
        self._sampler = sampler
        self._sampler_obs = sampler_obs
        self._sr = sr
        if sr is not None:
            self._sr.is_holomorphic = sampler.machine.is_holomorphic

        self._npar = self._machine.n_par

        self._batch_size = sampler.sample_shape[0]
        self._batch_size_obs = sampler.sample_shape[0]

        self.n_samples = n_samples
        self.n_discard = n_discard
        self.n_samples_obs = n_samples_obs
        self.n_discard_obs = n_discard_obs

        self._obs_samples_valid = False
        self._samples = _np.ndarray(
            (self._n_samples_node, self._batch_size, lindblad.hilbert.size)
        )
        self._samples_obs = _np.ndarray(
            (
                self._n_samples_obs_node,
                self._batch_size_obs,
                lindblad.hilbert.size_physical,
            )
        )

        self._der_logs = _np.ndarray(
            (self._n_samples_node, self._batch_size,
             self._npar), dtype=_np.complex128
        )

        self._der_loc_vals = _np.ndarray(
            (self._n_samples_node, self._batch_size,
             self._npar), dtype=_np.complex128
        )

        # Set the machine_pow of the sampler over the diagonal of the density matrix
        # to be |\rho(x,x)|
        sampler_obs.machine_pow = 1.0

    @property
    def n_samples(self):
        return self._n_samples

    @property
    def ldagl(self):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.
        """
        return self._loss_stats

    @property
    def n_samples_obs(self):
        return self._n_samples_obs

    @n_samples.setter
    def n_samples(self, n_samples):
        if n_samples <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples={}".format(n_samples)
            )
        self._n_samples = n_samples
        n_samples_chain = int(_np.ceil((n_samples / self._batch_size)))
        self._n_samples_node = int(_np.ceil(n_samples_chain / _MPI.size()))

    @n_samples_obs.setter
    def n_samples_obs(self, n_samples):
        if self._sampler_obs is None:
            return

        if n_samples <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples={}".format(n_samples)
            )
        self._n_samples_obs = n_samples
        n_samples_chain = int(_np.ceil((n_samples / self._batch_size_obs)))
        self._n_samples_obs_node = int(_np.ceil(n_samples_chain / _MPI.size()))

    @property
    def n_discard(self):
        return self._n_discard

    @property
    def n_discard_obs(self):
        return self._n_discard_obs

    @n_discard.setter
    def n_discard(self, n_discard):
        if n_discard is not None and n_discard < 0:
            raise ValueError(
                "Invalid number of discarded samples: n_discard={}".format(
                    n_discard)
            )
        self._n_discard = (
            n_discard
            if n_discard != None
            else self._n_samples_node * self._batch_size // 10
        )

    @n_discard_obs.setter
    def n_discard_obs(self, n_discard):
        if self._sampler_obs is None:
            return

        if n_discard is not None and n_discard < 0:
            raise ValueError(
                "Invalid number of discarded samples: n_discard={}".format(
                    n_discard)
            )
        self._n_discard_obs = (
            n_discard
            if n_discard is not None
            else self._n_samples_obs_node * self._batch_size_obs // 10
        )

    def _forward_and_backward(self):
        """
        Performs a number of VMC optimization steps.

        Args:
            n_steps (int): Number of steps to perform.
        """

        self._sampler.reset()
        self._obs_samples_valid = False

        # Burnout phase
        for _ in self._sampler.samples(self._n_discard):
            pass

        # Generate samples
        for i, sample in enumerate(self._sampler.samples(self._n_samples_node)):

            # Store the current sample
            self._samples[i] = sample

            # Compute Log derivatives
            self._der_logs[i] = self._machine.der_log(sample)

            self._der_loc_vals[i] = _der_local_values(
                self._lind, self._machine, sample, center_derivative=False
            )

        # flatten MC chain dimensions:
        self._der_logs = self._der_logs.reshape(-1, self._npar)

        # Estimate energy
        lloc, self._loss_stats = self._get_mc_superop_stats(self._lind)

        # Compute the (MPI-aware-)average of the derivatives
        der_logs_ave = _mean(self._der_logs, axis=0)

        # Center the log derivatives
        self._der_logs -= der_logs_ave

        # Compute the gradient
        grad = _covariance_sv(lloc, self._der_loc_vals, center_s=False)
        grad -= self._loss_stats.mean * der_logs_ave.conj()

        # Perform update
        if self._sr:
            dp = _np.empty(self._npar, dtype=_np.complex128)
            # flatten MC chain dimensions:
            derlogs = self._der_logs.reshape(-1, self._der_logs.shape[-1])
            self._sr.compute_update(derlogs, grad, dp)
            derlogs = self._der_logs.reshape(
                self._n_samples_node, self._batch_size, self._npar
            )
        else:
            dp = grad

        self._der_logs = self._der_logs.reshape(
            self._n_samples_node, self._batch_size, self._npar
        )

        return dp

    def sweep_diagonal(self):
        """
        Sweeps the diagonal of the density matrix with the observable sampler.
        """
        self._sampler_obs.reset()

        # Burnout phase
        for _ in self._sampler_obs.samples(self._n_discard_obs):
            pass

        # Generate samples
        for i, sample in enumerate(self._sampler_obs.samples(self._n_samples_obs_node)):

            # Store the current sample
            self._samples_obs[i] = sample

        self._obs_samples_valid = True

    def _estimate_stats(self, obs):
        return self._get_mc_obs_stats(obs)[1]

    def reset(self):
        self._sampler.reset()
        super().reset()

    def _get_mc_superop_stats(self, op):
        loc = _np.empty(self._samples.shape[0:2], dtype=_np.complex128)
        for i, sample in enumerate(self._samples):
            _local_values(op, self._machine, sample, out=loc[i])

        return loc, _statistics(_np.square(_np.abs(loc), dtype="complex128"))

    def _get_mc_obs_stats(self, op):
        if not self._obs_samples_valid:
            self.sweep_diagonal()

        loc = _local_values(op, self._machine, self._samples_obs)
        return loc, _statistics(loc)

    def __repr__(self):
        return "SteadyState(step_count={}, n_samples={}, n_discard={})".format(
            self.step_count, self.n_samples, self.n_discard
        )

    def info(self, depth=0):
        lines = [
            "{}: {}".format(name, info(obj, depth=depth + 1))
            for name, obj in [
                ("Liouvillian", self._lind),
                ("Machine", self._machine),
                ("Optimizer", self._optimizer_desc),
                ("SR solver", self._sr),
            ]
        ]
        return "\n  ".join([str(self)] + lines)



from .abstract_machine import AbstractMachine
import torch as _torch
import numpy as _np
import warnings


def _get_number_parameters(m):
    r"""Returns total number of variational parameters in a torch.nn.Module."""
    return sum(map(lambda p: p.numel(), _get_differentiable_parameters(m)))


def _get_differentiable_parameters(m):
    r"""Returns total number of variational parameters in a torch.nn.Module."""
    return filter(lambda p: p.requires_grad, m.parameters())


class Torch(AbstractMachine):
    def __init__(self, module, hilbert):
        self._module = _torch.jit.load(module) if isinstance(module, str) else module
        self._module.double()
        self._n_par = _get_number_parameters(self._module)
        self._parameters = list(_get_differentiable_parameters(self._module))
        self.n_visible = hilbert.size
        # TODO check that module has input shape compatible with hilbert size
        super().__init__(hilbert)

    @property
    def parameters(self):
        return (
            _torch.cat(
                tuple(p.view(-1) for p in _get_differentiable_parameters(self._module))
            )
            .detach()
            .numpy()
            .astype(_np.complex128)
        )

    def assign_beta(self, beta):
        self._module.beta = beta
        return

    def save(self, filename):
        _torch.save(self._module.state_dict(), filename)
        return

    def load(self, filename):
        self._module.load_state_dict(_torch.load(filename))
        return

    @parameters.setter
    def parameters(self, p):
        if not _np.all(p.imag == 0.0):
            warnings.warn(
                "PyTorch machines have real parameters, imaginary part will be discarded"
            )
        torch_pars = _torch.from_numpy(p.real)
        if torch_pars.numel() != self.n_par:
            raise ValueError(
                "p has wrong shape: {}; expected [{}]".format(
                    torch_pars.size(), self.n_par
                )
            )
        i = 0
        for x in map(
            lambda x: x.view(-1), _get_differentiable_parameters(self._module)
        ):
            x.data.copy_(torch_pars[i : i + len(x)].data)
            i += len(x)

    @property
    def n_par(self):
        r"""Returns the total number of trainable parameters in the machine.
        """
        return self._n_par

    def log_val(self, x, out=None):
        if len(x.shape) == 1:
            x = x[_np.newaxis, :]

        batch_shape = x.shape[:-1]

        with _torch.no_grad():
            t_out = self._module(_torch.from_numpy(x)).numpy().view(_np.complex128)

        if out is None:
            return t_out.reshape(batch_shape)

        _np.copyto(out, t_out.reshape(-1))

        return out

    def der_log(self, x, out=None):
        if len(x.shape) == 1:
            x = x[_np.newaxis, :]
        batch_shape = x.shape[:-1]
        x = x.reshape(-1, x.shape[-1])

        x = _torch.tensor(x, dtype=_torch.float64)
        out = x.new_empty([x.size(0), 2, self._n_par], dtype=self._parameters[0].dtype)
        m = self._module(x)

        for i in range(x.size(0)):
            dws_real = _torch.autograd.grad(
                m[i, 0], self._parameters, retain_graph=True
            )
            dws_imag = _torch.autograd.grad(
                m[i, 1], self._parameters, retain_graph=True
            )
            _torch.cat([dw.flatten() for dw in dws_real], out=out[i, 0, ...])
            _torch.cat([dw.flatten() for dw in dws_imag], out=out[i, 1, ...])

        out_complex = _np.zeros((out.size(0), out.size(2)), dtype=_np.complex128)
        out_complex = out[:, 0, :].numpy() + 1.0j * out[:, 1, :].numpy()

        return out_complex.reshape(
            tuple(list(batch_shape) + list(out_complex.shape[-1:]))
        )

    def vector_jacobian_prod(self, x, vec, out=None):

        if out is None:
            out = _np.empty(self.n_par, dtype=_np.complex128)

        def write_to(dst):
            dst = _torch.from_numpy(dst)
            i = 0
            for g in (
                p.grad.flatten() for p in self._module.parameters() if p.requires_grad
            ):
                dst[i : i + g.numel()].copy_(g)
                i += g.numel()

        def zero_grad():
            for g in (p.grad for p in self._module.parameters() if p.requires_grad):
                if g is not None:
                    g.zero_()

        vecj = _torch.empty(x.shape[0], 2, dtype=_torch.float64)

        def get_vec(is_real):
            if is_real:
                vecj[:, 0] = _torch.from_numpy(vec.real)
                vecj[:, 1] = _torch.from_numpy(vec.imag)
            else:
                vecj[:, 0] = _torch.from_numpy(vec.imag)
                vecj[:, 1] = _torch.from_numpy(-vec.real)
            return vecj

        y = self._module(_torch.from_numpy(x))
        zero_grad()
        y.backward(get_vec(True), retain_graph=True)
        write_to(out.real)
        zero_grad()
        y.backward(get_vec(False))
        write_to(out.imag)

        return out

    @property
    def is_holomorphic(self):
        r"""PyTorch models are real-valued only, thus non holomorphic.
        """
        return False

    @property
    def state_dict(self):
        from collections import OrderedDict

        return OrderedDict(
            [(k, v.detach().numpy()) for k, v in self._module.state_dict().items()]
        )


class TorchLogCosh(_torch.nn.Module):
    """
    Log(cosh) activation function for PyTorch modules
    """

    def __init__(self):
        """
        Init method.
        """
        super().__init__()  # init the base class

    def forward(self, input):
        """
        Forward pass of the function.
        """
        return -input + _torch.nn.functional.softplus(2.0 * input)


class TorchView(_torch.nn.Module):
    """
    Reshaping layer for PyTorch modules
    """

    def __init__(self, shape):
        """
        Init method.
        """
        super().__init__()  # init the base class
        self.shape = shape

    def forward(self, x):
        return x.view(*self.shape)



import numpy as _np
from netket import random as _random

from .abstract_sampler import AbstractSampler
from .metropolis_hastings import *
from .._C_netket import sampler as c_sampler

from numba import jit


class _hamiltonian_kernel:
    def __init__(self, hamiltonian):
        self._hamiltonian = hamiltonian
        self._sections = _np.empty(1, dtype=_np.int32)
        self._hamconn = self._hamiltonian.get_conn_flattened
        self._n_conn = self._hamiltonian.get_n_conn

    def apply(self, state, state_1, log_prob_corr):

        sections = self._sections
        sections = _np.empty(state.shape[0], dtype=_np.int32)
        vprimes = self._hamconn(state, sections)[0]

        self._choose(vprimes, sections, state_1, log_prob_corr)

        self._n_conn(state_1, sections)

        self._corr(sections, log_prob_corr)

    @staticmethod
    @jit(nopython=True)
    def _choose(states, sections, out, w):
        low_range = 0
        for i, s in enumerate(sections):
            n_rand = _random.randint(low_range, s)
            out[i] = states[n_rand]
            w[i] = _np.log(s - low_range)
            low_range = s

    @staticmethod
    @jit(nopython=True)
    def _corr(n_conn, w):
        for i, n in enumerate(n_conn):
            w[i] -= _np.log(n)


class MetropolisHamiltonian(AbstractSampler):
    """
    Sampling based on the off-diagonal elements of a Hamiltonian (or a generic Operator).
    In this case, the transition matrix is taken to be:

    .. math::
       T( \mathbf{s} \rightarrow \mathbf{s}^\prime) = \frac{1}{\mathcal{N}(\mathbf{s})}\theta(|H_{\mathbf{s},\mathbf{s}^\prime}|),

    where :math:`\theta(x)` is the Heaviside step function, and :math:`\mathcal{N}(\mathbf{s})`
    is a state-dependent normalization.
    The effect of this transition probability is then to connect (with uniform probability)
    a given state :math:`\mathbf{s}` to all those states :math:`\mathbf{s}^\prime` for which the Hamiltonian has
    finite matrix elements.
    Notice that this sampler preserves by construction all the symmetries
    of the Hamiltonian. This is in generally not true for the local samplers instead.
    """

    def __init__(
        self, machine, hamiltonian, n_chains=16, sweep_size=None, batch_size=None
    ):
        """
        Args:
           machine: A machine :math:`\Psi(s)` used for the sampling.
                    The probability distribution being sampled
                    from is :math:`F(\Psi(s))`, where the function
                    :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
           hamiltonian: The operator used to perform off-diagonal transition.
           n_chains: The number of Markov Chain to be run in parallel on a single process.
           sweep_size: The number of exchanges that compose a single sweep.
                       If None, sweep_size is equal to the number of degrees of freedom (n_visible).
           batch_size: The batch size to be used when calling log_val on the given Machine.
                       If None, batch_size is equal to the number Markov chains (n_chains).

       Examples:
           Sampling from a RBM machine in a 1D lattice of spin 1/2

           >>> import netket as nk
           >>>
           >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
           >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
           >>>
           >>> # RBM Spin Machine
           >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
           >>>
           >>> # Transverse-field Ising Hamiltonian
           >>> ha = nk.operator.Ising(hilbert=hi, h=1.0)
           >>>
           >>> # Construct a MetropolisHamiltonian Sampler
           >>> sa = nk.sampler.MetropolisHamiltonian(machine=ma,hamiltonian=ha)
        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.MetropolisHamiltonian(
                machine=machine,
                hamiltonian=hamiltonian,
                n_chains=n_chains,
                sweep_size=sweep_size,
                batch_size=batch_size,
            )
        else:
            self.sampler = PyMetropolisHastings(
                machine,
                _hamiltonian_kernel(hamiltonian),
                n_chains,
                sweep_size,
                batch_size,
            )
        super().__init__(machine, n_chains)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance


class MetropolisHamiltonianPt(AbstractSampler):
    """
     This sampler performs parallel-tempering
     moves in addition to the local moves implemented in `MetropolisLocal`.
     The number of replicas can be :math:`N_{\mathrm{rep}}` chosen by the user.
    """

    def __init__(self, machine, hamiltonian, n_replicas=16, sweep_size=None):
        """
        Args:
            machine: A machine :math:`\Psi(s)` used for the sampling.
                      The probability distribution being sampled
                      from is :math:`F(\Psi(s))`, where the function
                      :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
            hamiltonian: The operator used to perform off-diagonal transition.
            n_replicas: The number of replicas used for parallel tempering.
            sweep_size: The number of exchanges that compose a single sweep.
                         If None, sweep_size is equal to the number of degrees of freedom (n_visible).
        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.MetropolisHamiltonianPt(
                machine=machine,
                hamiltonian=hamiltonian,
                n_replicas=n_replicas,
                sweep_size=sweep_size,
            )
        else:
            raise ValueError(
                """Parallel Tempering samplers are not yet implemented
                for pure python machines"""
            )
        super().__init__(machine, 1)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance



import sys

import netket as _nk


def info(obj, depth=None):
    if hasattr(obj, "info"):
        return obj.info(depth)
    else:
        return str(obj)


def make_optimizer_fn(arg, ma):
    """
    Utility function to create the optimizer step function for VMC drivers.

    It currently supports three kinds of inputs:

    1. A NetKet optimizer, i.e., a subclass of `netket.optimizer.Optimizer`.

    2. A 3-tuple (init, update, get) of optimizer functions as used by the JAX
       optimizer submodule (jax.experimental.optimizers).

       The update step p0 -> p1 with bare step dp is computed as
            x0 = init(p0)
            x1 = update(i, dp, x1)
            p1 = get(x1)

    3. A single function update with signature p1 = update(i, dp, p0) returning the
       updated parameter value.
    """
    if isinstance(arg, tuple) and len(arg) == 3:
        init, update, get = arg

        def optimize_fn(i, grad, p):
            x0 = init(p)
            x1 = update(i, grad, x0)
            return get(x1)

        desc = "JAX-like optimizer"
        return optimize_fn, desc

    elif issubclass(type(arg), _nk.optimizer.Optimizer):

        arg.init(ma.n_par, ma.is_holomorphic)

        def optimize_fn(_, grad, p):
            arg.update(grad, p)
            return p

        desc = info(arg)
        return optimize_fn, desc

    elif callable(arg):
        import inspect

        sig = inspect.signature(arg)
        if not len(sig.parameters) == 3:
            raise ValueError(
                "Expected netket.optimizer.Optimizer subclass, JAX optimizer, "
                + " or callable f(i, grad, p); got callable with signature {}".format(
                    sig
                )
            )
        desc = "{}{}".format(arg.__name__, sig)
        return arg, desc
    else:
        raise ValueError(
            "Expected netket.optimizer.Optimizer subclass, JAX optimizer, "
            + " or callable f(i, grad, p); got {}".format(arg)
        )


def tree_map(fun, tree):

    if tree is None:
        result = None

    elif "_C_netket" in str(type(tree)):
        result = fun(tree)

    elif type(tree) == list:
        result = []
        for val in tree:
            result.append(fun, tree_map(val))

    else:
        result = {}
        for key in tree:
            result[key] = tree_map(fun, tree[key])

    return result



from ._C_netket.hilbert import *


def Qubit(graph):
    """
    Constructs a new ``Qubit`` given a graph.

    Args:
        graph: Graph representation of sites.

    Examples:
        Simple qubit hilbert space.

        >>> from netket.graph import Hypercube
        >>> from netket.hilbert import Qubit
        >>> g = Hypercube(length=10,n_dim=2,pbc=True)
        >>> hi = Qubit(graph=g)
        >>> print(hi.size)
        100
    """
    return CustomHilbert(graph, local_states=[0, 1])



from ._vmc import Vmc as _Vmc
from ._C_netket.optimizer import SR as _SR
import warnings
from netket.vmc_common import tree_map


def Vmc(
    hamiltonian,
    sampler,
    optimizer,
    n_samples,
    discarded_samples=None,
    discarded_samples_on_init=0,
    target="energy",
    method="Sr",
    diag_shift=0.01,
    use_iterative=False,
    use_cholesky=None,
    sr_lsq_solver="LLT",
):

    if use_cholesky and sr_lsq_solver != "LLT":
        raise ValueError(
            "Inconsistent options specified: `use_cholesky && sr_lsq_solver != 'LLT'`."
        )

    if discarded_samples_on_init != 0:
        warnings.warn(
            "discarded_samples_on_init does not have any effect and should not be used",
            DeprecationWarning,
        )

    warnings.warn(
        "netket.variational.Vmc will be deprecated in version 3, use netket.Vmc instead",
        PendingDeprecationWarning,
    )

    if method == "Gd":
        return _Vmc(
            hamiltonian=hamiltonian,
            sampler=sampler,
            optimizer=optimizer,
            n_samples=n_samples,
            n_discard=discarded_samples,
            sr=None,
        )
    elif method == "Sr":
        sr = _SR(
            lsq_solver=sr_lsq_solver,
            diag_shift=diag_shift,
            use_iterative=use_iterative,
            is_holomorphic=sampler.machine.is_holomorphic,
        )
        return _Vmc(
            hamiltonian=hamiltonian,
            sampler=sampler,
            optimizer=optimizer,
            n_samples=n_samples,
            n_discard=discarded_samples,
            sr=sr,
        )
    else:
        raise ValueError("Allowed method options are Gd and Sr")


# Higher-level VMC functions:


def estimate_expectations(
    ops, sampler, n_samples, n_discard=None, compute_gradients=False
):
    """
    For a sequence of linear operators, computes a statistical estimate of the
    respective expectation values, variances, and optionally gradients of the
    expectation values with respect to the variational parameters.

    The estimate is based on `n_samples` configurations
    obtained from `sampler`.

    Args:
        ops: pytree of linear operators
        sampler: A NetKet sampler
        n_samples: Number of MC samples used to estimate expectation values
        n_discard: Number of MC samples dropped from the start of the
            chain (burn-in). Defaults to `n_samples //10`.
        compute_gradients: Whether to compute the gradients of the
            observables.

    Returns:
        Either `stats` or, if `der_logs` is passed, a tuple of `stats` and `grad`:
            stats: A sequence of Stats object containing mean, variance,
                and MC diagonstics for each operator in `ops`.
            grad: A sequence of gradients of the expectation value of `op`,
                  as ndarray of shape `(psi.n_par,)`, for each `op` in `ops`.
    """

    from netket.operator import local_values as _local_values
    from ._C_netket import stats as nst

    psi = sampler.machine

    if not n_discard:
        n_discard = n_samples // 10

    # Burnout phase
    sampler.generate_samples(n_discard)
    # Generate samples
    samples = sampler.generate_samples(n_samples)

    if compute_gradients:
        der_logs = psi.der_log(samples)

    def estimate(op):
        lvs = _local_values(op, psi, samples)
        stats = nst.statistics(lvs)

        if compute_gradients:
            grad = nst.covariance_sv(lvs, der_logs)
            return stats, grad
        else:
            return stats

    return tree_map(estimate, ops)



import abc
import numpy as _np


class AbstractMachine(abc.ABC):
    """Abstract class for NetKet machines"""

    def __init__(self, hilbert):
        super().__init__()
        self.hilbert = hilbert

    @abc.abstractmethod
    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `x` and stores the result into `out`.

        Args:
            x: Either a
                * vector of `float64` of size `self.n_visible` or
                * a matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. If `x` is a matrix then
                length of `out` should be `x.shape[0]`. If `x` is a vector,
                then length of `out` should be 1.

        Returns:
            A complex number when `x` is a vector and vector when `x` is a
            matrix.
            """
        pass

    @abc.abstractmethod
    def vector_jacobian_prod(selv, x, vec, out=None):
        r"""Computes the scalar product between gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and a vector `vec`. The result is stored into `out`.

        Args:
             x: a matrix of `float64` of shape `(*, self.n_visible)`.
             vec: a `complex128` vector used to compute the inner product with the jacobian.
             out: The result of the inner product, it is a vector of `complex128` and length `self.n_par`.


        Returns:
             `out`
        """
        return NotImplementedError

    def jacobian_vector_prod(self, v, vec, out=None):
        return NotImplementedError

    def der_log(self, x, out=None):
        r"""Computes the gradient of the logarithm of the wavefunction for a
        batch of visible configurations `x` and stores the result into `out`.

        Args:
            x: Either a
                * vector of `float64` of size `self.n_visible` or
                * a matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`. If `x` is a matrix then of
                `out` should be a matrix of shape `(v.shape[0], self.n_par)`.
                If `x` is a vector, then `out` should be a vector of length
                `self.n_par`.

        Returns:
            `out`
            """
        return NotImplementedError

    def to_array(self, normalize=True, b_size=512):
        if self.hilbert.is_indexable:
            all_psis = _np.zeros(self.hilbert.n_states, dtype=_np.complex128)
            batch_states = _np.zeros((b_size, self.hilbert.size))
            it = self.hilbert.states().__iter__()
            for i in range(self.hilbert.n_states // b_size + 1):
                for j in range(b_size):
                    try:
                        batch_states[j] = next(it)
                    except StopIteration:
                        batch_states.resize(j, self.hilbert.size)
                        break
                all_psis[
                    i * b_size : i * b_size + batch_states.shape[0]
                ] = self.log_val(batch_states)

                logmax = _np.max(all_psis.real)
                all_psis = _np.exp(all_psis - logmax)

            if normalize:
                norm = _np.linalg.norm(all_psis)
                all_psis = all_psis / norm

            return all_psis
        else:
            return AssertionError

    @property
    @abc.abstractmethod
    def is_holomorphic(self):
        r"""Returns whether the wave function is holomorphic."""
        return NotImplementedError

    @property
    @abc.abstractmethod
    def n_par(self):
        r"""The number of variational parameters in the machine."""
        return NotImplementedError

    @property
    @abc.abstractmethod
    def state_dict(self):
        r"""A dictionary containing the parameters of this machine"""
        return NotImplementedError

    @property
    def parameters(self):
        return _np.concatenate(tuple(p.reshape(-1) for p in self.state_dict.values()))

    @parameters.setter
    def parameters(self, p):
        if p.shape != (self.n_par,):
            raise ValueError(
                "p has wrong shape: {}; expected ({},)".format(
                    p.shape, self.n_par)
            )
        i = 0
        for x in map(lambda x: x.reshape(-1), self.state_dict.values()):
            _np.copyto(x, p[i: i + x.size])
            i += x.size

    def save(self, file):
        _np.save(file, self.parameters, allow_pickle=False)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools as _itertools

from . import _core
from ._C_netket.exact import *
import numpy as np
from scipy.sparse.linalg import LinearOperator, bicgstab


def _ExactTimePropagation_iter(self, dt, n_iter=None):
    """
    iter(self: ExactTimePropagation, dt: float, n_iter: int=None) -> int

    Returns a generator which advances the time evolution by dt,
    yielding after every step.

    Args:
        dt (float): The size of the time step.
        n_iter (int=None): The number of steps or None, for no limit.

    Yields:
        int: The current step.
    """
    for i in _itertools.count():
        if n_iter and i >= n_iter:
            return
        self.advance(dt)
        yield i


class EdResult(object):
    def __init__(self, eigenvalues, eigenvectors):
        # NOTE: These conversions are required because our old C++ code stored
        # eigenvalues and eigenvectors as Python lists :(
        self._eigenvalues = eigenvalues.tolist()
        self._eigenvectors = (
            [eigenvectors[:, i] for i in range(eigenvectors.shape[1])]
            if eigenvectors is not None
            else []
        )

    @property
    def eigenvalues(self):
        r"""Eigenvalues of the Hamiltonian.
        """
        return self._eigenvalues

    @property
    def eigenvectors(self):
        r"""Eigenvectors of the Hamiltonian.
        """
        return self._eigenvectors

    def mean(self, operator, which):
        import numpy

        x = self._eigenvectors[which]
        return numpy.vdot(x, operator(x))


def lanczos_ed(
    operator,
    matrix_free=False,
    first_n=1,
    max_iter=1000,
    seed=None,
    precision=1e-14,
    compute_eigenvectors=False,
):
    r"""Computes `first_n` smallest eigenvalues and, optionally, eigenvectors
    of a Hermitian operator using the Lanczos method.

    Args:
        operator: The operator to diagnolize.
        matrix_free: If true, matrix elements are computed on the fly.
            Otherwise, the operator is first converted to a sparse matrix.
        first_n: The number of eigenvalues to compute.
        max_iter: The maximum number of iterations.
        seed: **Ignored**. Accepted for backward compatibility only.
        precision: The precision to which the eigenvalues will be
            computed.
        compute_eigenvectors: Whether or not to return the
            eigenvectors of the operator. With ARPACK, not requiring the
            eigenvectors has almost no performance benefits.


    Examples:
        Testing the number of eigenvalues saved when solving a simple
        1D Ising problem.

        ```python
        >>> import netket as nk
        >>> hilbert = nk.hilbert.Spin(
        ...     nk.graph.Hypercube(length=8, n_dim=1, pbc=True), s=0.5)
        >>> hamiltonian = nk.operator.Ising(h=1.0, hilbert=hilbert)
        >>> r = nk.exact.lanczos_ed(
        ...     hamiltonian, first_n=3, compute_eigenvectors=True)
        >>> r.eigenvalues
        [-10.251661790966047, -10.054678984251746, -8.690939214837037]
        ```

    """
    from scipy.sparse.linalg import eigsh

    result = eigsh(
        operator.to_linear_operator() if matrix_free else operator.to_sparse(),
        k=first_n,
        which="SA",
        maxiter=max_iter,
        tol=precision,
        return_eigenvectors=compute_eigenvectors,
    )
    if compute_eigenvectors:
        return EdResult(*result)
    return EdResult(result, None)


def full_ed(operator, first_n=1, compute_eigenvectors=False):
    r"""Computes `first_n` smallest eigenvalues and, optionally, eigenvectors
    of a Hermitian operator by full diagonalization.

    Args:
        operator: Operator to diagnolize.
        first_n: (Deprecated) Number of eigenvalues to compute.
            This has no performance impact, as full_ed will compute all
            eigenvalues anyway.
        compute_eigenvectors: Whether or not to return the
            eigenvectors of the operator.

    Examples:
        Testing the numer of eigenvalues saved when solving a simple
        1D Ising problem.

        ```python
        >>> import netket as nk
        >>> hilbert = nk.hilbert.Spin(
        ...     nk.graph.Hypercube(length=8, n_dim=1, pbc=True), s=0.5)
        >>> hamiltonian = nk.operator.Ising(h=1.0, hilbert=hilbert)
        >>> r = nk.exact.lanczos_ed(
        ...     hamiltonian, first_n=3, compute_eigenvectors=True)
        >>> len(r.eigenvalues)
        3
        ```
    """
    from numpy.linalg import eigh, eigvalsh

    dense_op = operator.to_dense()

    if not (1 <= first_n < dense_op.shape[0]):
        raise ValueError("first_n must be in range 1..dim(operator)")

    if compute_eigenvectors:
        w, v = eigh(dense_op)
        return EdResult(w[:first_n], v[:, :first_n])
    else:
        w = eigvalsh(dense_op)
        return EdResult(w[:first_n], None)


ExactTimePropagation.iter = _ExactTimePropagation_iter


def steady_state(lindblad, sparse=False, method="ed", rho0=None, **kwargs):
    r"""Computes the numerically exact steady-state of a lindblad master equation.
    The computation is performed either through the exact diagonalization of the
    hermitian L^\dagger L matrix, or by means of an iterative solver (bicgstabl)
    targeting the solution of the non-hermitian system L\rho = 0 && \Tr[\rho] = 1.

    Note that for systems with 7 or more sites it is usually computationally impossible
    to build the full lindblad operator and therefore only `iterative` will work.

    Note that for systems with hilbert spaces with dimensions above 40k, tol
    should be set to a lower value if the steady state has non-trivial correlations.

    Args:
        lindblad: The lindbladian encoding the master equation.
        sparse: Whever to use sparse matrices (default: False)
        method: 'ed' (exact diagonalization) or 'iterative' (iterative bicgstabl)
        rho0: starting density matrix for the iterative diagonalization (default: None)
        kwargs...: additional kwargs passed to bicgstabl

    Optional args for iterative:
        For full docs please consult SciPy documentation at
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.bicgstab.html

        maxiter: maximum number of iterations for the iterative solver (default: None)
        tol: The precision for the calculation (default: 1e-05)
        callback: User-supplied function to call after each iteration. It is called as callback(xk),
                  where xk is the current solution vector

    """
    from numpy import sqrt, matrix

    M = lindblad.hilbert.hilbert_physical.n_states

    if method == "ed":
        if not sparse:
            from numpy.linalg import eigh

            lind_mat = matrix(lindblad.to_dense())

            ldagl = lind_mat.H * lind_mat
            w, v = eigh(ldagl)

        else:
            from scipy.sparse.linalg import eigsh

            lind_mat = lindblad.to_sparse()
            ldagl = lind_mat.H * lind_mat

            w, v = eigsh(ldagl, which="SM", k=2)

        print("Minimum eigenvalue is: ", w[0])
        rho = matrix(v[:, 0].reshape((M, M)))
        rho = rho / rho.trace()

    elif method == "iterative":

        iHnh = -1j * lindblad.get_effective_hamiltonian()
        if sparse:
            iHnh = iHnh.to_sparse()
            J_ops = [j.to_sparse() for j in lindblad.jump_ops]
        else:
            iHnh = iHnh.to_dense()
            J_ops = [j.to_dense() for j in lindblad.jump_ops]

        # This function defines the product Liouvillian x densitymatrix, without
        # constructing the full density matrix (passed as a vector M^2).

        # An extra row is added at the bottom of the therefore M^2+1 long array,
        # with the trace of the density matrix. This is needed to enforce the
        # trace-1 condition.

        # The logic behind the use of Hnh_dag_ and Hnh_ is derived from the
        # convention adopted in local_liouvillian.cc, and inspired from reference
        # arXiv:1504.05266
        def matvec(rho_vec):
            rho = rho_vec[:-1].reshape((M, M))

            out = np.zeros((M ** 2 + 1), dtype="complex128")
            drho = out[:-1].reshape((M, M))

            drho += rho @ iHnh + iHnh.conj().T @ rho
            for J in J_ops:
                drho += (J @ rho) @ J.conj().T

            out[-1] = rho.trace()
            return out

        L = LinearOperator((M ** 2 + 1, M ** 2 + 1), matvec=matvec)

        # Initial density matrix ( + trace condition)
        Lrho_start = np.zeros((M ** 2 + 1), dtype="complex128")
        if rho0 is None:
            Lrho_start[0] = 1.0
            Lrho_start[-1] = 1.0
        else:
            Lrho_start[:-1] = rho0.reshape(-1)
            Lrho_start[-1] = rho0.trace()

        # Target residual (everything 0 and trace 1)
        Lrho_target = np.zeros((M ** 2 + 1), dtype="complex128")
        Lrho_target[-1] = 1.0

        # Iterative solver
        print("Starting iterative solver...")
        res, info = bicgstab(L, Lrho_target, x0=Lrho_start, **kwargs)

        rho = res[1:].reshape((M, M))
        if info == 0:
            print("Converged trace residual is ", res[-1])
        elif info > 0:
            print(
                "Failed to converge after ", info, " ( traceresidual is ", res[-1], " )"
            )
        elif info < 0:
            print("An error occured: ", info)

    else:
        raise ValueError("method must be 'ed'")

    return rho


@_core.deprecated(
    "`ImagTimePropagation` is deprecated. Please use "
    '`ExactTimePropagation(..., propagation_type="imaginary")` instead.'
)
def ImagTimePropagation(*args, **kwargs):
    """
    Returns `ExactTimePropagation(..., propagation_type="imaginary")` for
    backwards compatibility.

    Deprecated (NetKet 2.0): Use `ExactTimePropagation` directly.
    """
    kwargs["propagation_type"] = "imaginary"
    return ExactTimePropagation(*args, **kwargs)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "dynamics",
    "exact",
    "graph",
    "hilbert",
    "layer",
    "machine",
    "operator",
    "optimizer",
    "output",
    "random",
    "sampler",
    "stats",
    "supervised",
    "utils",
    "variational",
]

from . import (
    _C_netket,
    dynamics,
    exact,
    graph,
    hilbert,
    layer,
    logging,
    machine,
    operator,
    optimizer,
    output,
    random,
    sampler,
    stats,
    supervised,
    utils,
    variational,
    _vmc,
    _steadystate,
)
from ._C_netket import MPI

# Main applications
from ._vmc import Vmc
from ._qsr import Qsr
from ._steadystate import SteadyState



import abc

from netket._core import deprecated, warn_deprecation
import netket as _nk

from netket.logging import JsonLog as _JsonLog

from netket.vmc_common import make_optimizer_fn, tree_map

from tqdm import tqdm

import warnings


def _obs_stat_to_dict(value):
    st = value.asdict()
    st["Mean"] = st["Mean"].real
    return st


# Note: to implement a new Driver (see also _vmc.py for an example)
# If you want to inherit the nice interface of AbstractMCDriver, you should
# subclass it, defining the following methods:
# - Either _forward_and_backward or individually _forward, _backward, that should
#   compute the loss function and the gradient. If the driver is minimizing or
#   maximising some loss function, this quantity should be assigned to self._stats
#   in order to monitor it.
# - _estimate_stats should return the MC estimate of a single operator
# - reset should reset the driver (usually the sampler).
# - info should return a string with an overview of the driver.
# - The __init__ method shouldbe called with the machine and the optimizer. If this
#   driver is minimising a loss function and you want it's name to show up automatically
#   in the progress bar/ouput files you should pass the optional keyword argument
#   minimized_quantity_name.
class AbstractVariationalDriver(abc.ABC):
    """Abstract base class for NetKet Variational Monte Carlo drivers"""

    def __init__(self, machine, optimizer, minimized_quantity_name=""):
        self._mynode = _nk.MPI.rank()
        self._obs = {}  # to deprecate
        self._loss_stats = None
        self._loss_name = minimized_quantity_name
        self.step_count = 0

        self._machine = machine
        self._optimizer_step, self._optimizer_desc = make_optimizer_fn(
            optimizer, self._machine
        )

    def _forward_and_backward(self):
        """
        Performs the forward and backward pass at the same time.
        Concrete drivers should either override this method, or override individually
        _forward and _backward.

        :return: the update for the weights.
        """
        self._forward()
        dp = self._backward()
        return dp

    def _forward(self):
        """
        Performs the forward pass, computing the loss function.
        Concrete should either implement _forward and _backward or the joint method
        _forward_and_backward.
        """
        raise NotImplementedError()

    def _backward(self):
        """
        Performs the backward pass, computing the update for the parameters.
        Concrete should either implement _forward and _backward or the joint method
        _forward_and_backward.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _estimate_stats(self, observable):
        """
        Returns the MCMC statistics for the expectation value of an observable.
        Must be implemented by super-classes of AbstractVMC.

        :param observable: A quantum operator (netket observable)
        :return:
        """
        pass

    @abc.abstractmethod
    def reset(self):
        """
        Resets the driver.
        Concrete drivers should also call super().reset() to ensure that the step
        count is set to 0.
        """
        self.step_count = 0
        pass

    @abc.abstractmethod
    def info(self, depth=0):
        """
        Returns an info string used to print information to screen about this driver.
        """
        pass

    @property
    def machine(self):
        """
        Returns the machine that is optimized by this driver.
        """
        return self._machine

    def iter(self, n_steps, step=1):
        """
        Returns a generator which advances the VMC optimization, yielding
        after every `step_size` steps.

        Args:
            :n_iter (int=None): The total number of steps to perform.
            :step_size (int=1): The number of internal steps the simulation
                is advanced every turn.

        Yields:
            int: The current step.
        """
        for _ in range(0, n_steps, step):
            for i in range(0, step):
                dp = self._forward_and_backward()
                if i == 0:
                    yield self.step_count

                self.update_parameters(dp)

    def advance(self, steps=1):
        """
        Performs `steps` optimization steps.

        :param steps: (Default=1) number of steps
        """
        for _ in self.iter(steps):
            pass

    def run(
        self,
        n_iter,
        out=None,
        obs=None,
        show_progress=True,
        save_params_every=50,  # for default logger
        write_every=50,  # for default logger
        step_size=1,  # for default logger
        output_prefix=None,  # TODO: deprecated
    ):
        """
        Executes the Monte Carlo Variational optimization, updating the weights of the network
        stored in this driver for `n_iter` steps and dumping values of the observables `obs`
        in the output `logger`. If no logger is specified, creates a json file at `output_prefix`,
        overwriting files with the same prefix.

        !! Compatibility v2.1
            Before v2.1 the order of the first two arguments, `n_iter` and `output_prefix` was
            reversed. The reversed ordering will still be supported until v3.0, but is deprecated.

        Args:
            :n_iter: the total number of iterations
            :out: A logger object to be used to store simulation log and data.
                If this argument is a string, it will be used as output prefix for the standard JSON logger.
            :obs: An iterable containing all observables that should be computed
            :save_params_every: Every how many steps the parameters of the network should be
            serialized to disk (ignored if logger is provided)
            :write_every: Every how many steps the json data should be flushed to disk (ignored if
            logger is provided)
            :step_size: Every how many steps should observables be logged to disk (default=1)
            :show_progress: If true displays a progress bar (default=True)
            :output_prefix: (Deprecated) The prefix at which json output should be stored (ignored if out
              is provided).
        """

        # TODO Remove this deprecated code in v3.0
        # manage deprecated where argument names are not specified, and
        # prefix is passed as the first positional argument and the number
        # of iterations as a second argument.
        if type(n_iter) is str and type(out) is int:
            n_iter, out = out, n_iter
            warn_deprecation(
                "The positional syntax run(output_prefix, n_iter, **args) is deprecated, use run(n_iter, output_prefix, **args) instead."
            )

        if obs is None:
            # TODO remove the first case after deprecation of self._obs in 3.0
            if len(self._obs) != 0:
                obs = self._obs
            else:
                obs = {}

        # output_prefix is deprecated. out should be used and takes over
        # error out if both are passed
        # TODO: remove in v3.0
        if out is not None and output_prefix is not None:
            raise ValueError(
                "Invalid out and output_prefix arguments. Only one of the two can be passed. Note that output_prefix is deprecated and you should use out."
            )
        elif out is None and output_prefix is not None:
            warn_deprecation(
                "The output_prefix argument is deprecated. Use out instead."
            )
            out = output_prefix

        if out is None:
            print(
                "No output specified (out=[apath|nk.logging.JsonLogger(...)])."
                "Running the optimization but not saving the output."
            )

        # Log only non-root nodes
        if self._mynode == 0:
            # if out is a path, create an overwriting Json Log for output
            if isinstance(out, str):
                logger = _JsonLog(out, "w", save_params_every, write_every)
            else:
                logger = out
        else:
            logger = None

        with tqdm(
            self.iter(n_iter, step_size), total=n_iter, disable=not show_progress
        ) as itr:
            for step in itr:
                # if the cost-function is defined then report it in the progress bar
                if self._loss_stats is not None:
                    itr.set_postfix_str(self._loss_name + "=" + str(self._loss_stats))

                obs_data = self.estimate(obs)

                if self._loss_stats is not None:
                    obs_data[self._loss_name] = self._loss_stats

                log_data = tree_map(_obs_stat_to_dict, obs_data)

                if logger is not None:
                    logger(step, log_data, self.machine)

        # flush at the end of the evolution so that final values are saved to
        # file
        if logger is not None:
            logger.flush(self.machine)

    def estimate(self, observables):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.

        Args:
            observables: A pytree of operators for which statistics should be computed.

        Returns:
            A pytree of the same structure as the input, containing MCMC statistics
            for the corresponding operators as leaves.
        """
        return tree_map(self._estimate_stats, observables)

    def update_parameters(self, dp):
        """
        Updates the parameters of the machine using the optimizer in this driver

        Args:
            :param dp: the gradient
        """
        self._machine.parameters = self._optimizer_step(
            self.step_count, dp, self._machine.parameters
        )
        self.step_count += 1

    @deprecated()
    def add_observable(self, obs, name):
        """
        Add an observables to the set of observables that will be computed by default
        in get_obervable_stats.

        This function is deprecated in favour of `estimate`.

        Args:
            obs: the operator encoding the observable
            name: a string, representing the name of the observable
        """
        self._obs[name] = obs

    @deprecated()
    def get_observable_stats(self, observables=None, include_energy=True):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.

        Args:
            observables: A dictionary of the form {name: observable} or a list
                of tuples (name, observable) for which statistics should be computed.
                If observables is None or not passed, results for those observables
                added to the driver by add_observables are computed.
            include_energy: Whether to include the energy estimate (which is already
                computed as part of the VMC step) in the result.

        Returns:
            A dictionary of the form {name: stats} mapping the observable names in
            the input to corresponding Stats objects.

            If `include_energy` is true, then the result will further contain the
            energy statistics with key "Energy".
        """
        if not observables:
            observables = self._obs

        if self._loss_name is None:
            include_energy = False

        result = self.estimate(observables)

        if include_energy:
            result[self._loss_name] = self._loss_stats

        return result



from .abstract_sampler import AbstractSampler
from .metropolis_hastings import *
from .metropolis_local import MetropolisLocal, MetropolisLocalPt
from .metropolis_exchange import MetropolisExchange, MetropolisExchangePt
from .metropolis_hamiltonian import MetropolisHamiltonian, MetropolisHamiltonianPt
from .custom_sampler import CustomSampler, CustomSamplerPt
from .exact_sampler import *



import json as _json
from os import path as _path


def _exists_json(prefix):
    return _path.exists(prefix + ".log") or _path.exists(prefix + ".wf")


class JsonLog:
    """
    Creates a Json Logger sink object, that can be passed with keyword argument `logger` to Monte
    Carlo drivers in order to serialize the outpit data of the simulation.

    Args:
        output_prefix: the name of the output files before the extension
        save_params_every: every how many iterations should machine parameters be flushed to file
        write_every: every how many iterations should data be flushed to file
        mode: Specify the behaviour in case the file already exists at this output_prefix. Options
        are
        - `[w]rite`: (default) overwrites file if it already exists;
        - `[a]ppend`: appends to the file if it exists, overwise creates a new file;
        - `[x]` or `fail`: fails if file already exists;
    """

    def __init__(
        self, output_prefix, mode="write", save_params_every=50, write_every=50
    ):
        # Shorthands for mode
        if mode == "w":
            mode = "write"
        elif mode == "a":
            mode = "append"
        elif mode == "x":
            mode = "fail"

        if not (
            (mode == "write")
            or (mode == "append")
            or (mode == "fail")
        ):
            raise ValueError(
                "Mode not recognized: should be one of `[w]rite`, `[a]ppend` or `[x]`(fail)."
            )

        file_exists = _exists_json(output_prefix)

        starting_json_content = {"Output": []}

        if file_exists and mode == "append":
            # if there is only the .wf file but not the json one, raise an error
            if not _path.exists(output_prefix + ".log"):
                raise ValueError(
                    "History file does not exists, but wavefunction file does. Please change `output_prefix or set mode=`write`."
                )

            starting_json_content = _json.load(open(output_prefix + ".log"))

        elif file_exists and mode == "fail":
            raise ValueError(
                "Output file already exists. Either delete it manually or change `output_prefix`."
            )

        self._json_out = starting_json_content
        self._prefix = output_prefix
        self._write_every = write_every
        self._save_params_every = save_params_every
        self._old_step = 0

    def __call__(self, step, item, machine):
        item["Iteration"] = step

        self._json_out["Output"].append(item)

        if step % self._write_every == 0 or step == self._old_step - 1:
            self._flush_log()
        if step % self._save_params_every == 0 or step == self._old_step - 1:
            self._flush_params(machine)

        self._old_step = step

    def _flush_log(self):
        with open(self._prefix + ".log", "w") as outfile:
            _json.dump(self._json_out, outfile)

    def _flush_params(self, machine):
        machine.save(self._prefix + ".wf")

    def flush(self, machine=None):
        """
        Writes to file the content of this logger.

        :param machine: optionally also writes the parameters of the machine.
        """
        self._flush_log()

        if machine is not None:
            self._flush_params(machine)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from collections import OrderedDict
from functools import reduce
from pickle import dump, load
import os

import numpy as np

os.environ["JAX_ENABLE_X64"] = "1"
import jax

from .cxx_machine import CxxMachine

__all__ = ["Jax"]


class Jax(CxxMachine):
    def __init__(self, hilbert, module, seed=None):
        """
        Wraps a stax network (which is a tuple of `init_fn` and `predict_fn`)
        so that it can be used as a NetKet machine.

        Args:
            hilbert: Hilbert space on which the state is defined. Should be a
                subclass of `netket.hilbert.Hilbert`.
            module: A pair `(init_fn, predict_fn)`. See the documentation of
                `jax.experimental.stax` for more info.
            seed: Seed to use to construct `jax.random.PRNGKey` for
                initialisation of parameters. If `None` seed will be chosen
                automatically (which is __not__ synchronized between MPI
                processes so prefer `self.init_random_parameters` if you're
                using multiple MPI tasks)
        """
        # NOTE: The following call to __init__ is important!
        super(Jax, self).__init__(hilbert)
        if seed is None:
            # NOTE(twesterhout): I believe, jax uses 32-bit integers internally
            # to represent the seed. Hence the limit
            seed = np.random.randint(2 ** 32 - 1)
        init_fn, self._forward_fn = module
        self._forward_fn = jax.jit(self._forward_fn)
        input_shape = (-1, hilbert.size)
        output_shape, self._params = init_fn(jax.random.PRNGKey(seed), input_shape)
        if output_shape != (-1, 2):
            raise ValueError("module's output_shape is weird, check your network")
        # Computes total number of parameters
        n_par = sum(reduce(lambda n, p: n + p.size, layer, 0) for layer in self._params)
        # Save the initial dtype (for debugging mostly)
        dtypes = set(p.dtype for layer in self._params for p in layer)
        if len(dtypes) != 1:
            raise ValueError("module parameters have different dtypes")
        self._dtype = next(iter(dtypes))
        self._complex_dtype = {
            np.dtype("float32"): np.complex64,
            np.dtype("float64"): np.complex128,
        }[self._dtype]
        assert all(
            np.asarray(p).flags.c_contiguous for layer in self._params for p in layer
        ), "sorry, column major order is not supported (yet)"
        self._n_par = lambda: n_par
        # Computes the Jacobian matrix using backprop
        self._jacobian = jax.jacrev(self._forward_fn)
        self._jacobian = jax.jit(self._jacobian)

    def _log_val(self, x, out):
        """
        Do not use this function directly! It's an implementation detail!

        See `self.log_val`.
        """
        out[:] = (
            np.asarray(self._forward_fn(self._params, x))
            .view(dtype=self._complex_dtype)
            .squeeze()
        )

    def _der_log(self, x, out):
        """
        Do not use this function directly! It's an implementation detail!

        See `self.log_val`.
        """
        J = self._jacobian(self._params, x)
        batch_size = x.shape[0]
        i = 0
        for g in (g.reshape(batch_size, 2, -1) for layer in J for g in layer):
            n = g.shape[2]
            out[:, i : i + n].real = g[:, 0, :]
            out[:, i : i + n].imag = g[:, 1, :]
            i += n

    def _is_holomorphic(self):
        return False

    def _set_parameters(self, new_parameters):
        state = []
        i = 0
        for layer in self._params:
            layer_state = ()
            for p in layer:
                if not np.all(p.imag == 0):
                    raise ValueError("parameters are purely real")
                # NOTE: This relies on the fact that all our parameters are
                # stored in row major order
                layer_state += (
                    np.ascontiguousarray(
                        new_parameters[i : i + p.size].real.reshape(p.shape)
                    ),
                )
                i += p.size
            state.append(layer_state)
        self._params = state

    @property
    def dtype(self):
        """
        Datatype of the parameters.
        """
        return self._dtype

    def state_dict(self):
        state = []
        for i, layer in enumerate(self._params):
            for j, p in enumerate(layer):
                state.append((str((i, j)), np.asarray(p).view()))
        return OrderedDict(state)

    def load_state_dict(self, other):
        state = []
        for i, layer in enumerate(self._params):
            layer_state = ()
            for j, p in enumerate(layer):
                name = str((i, j))
                if name not in other:
                    raise ValueError(
                        "state is missing required field {!r}".format(name)
                    )
                value = other[name]
                if p.shape != value.shape:
                    raise ValueError(
                        "field {!r} has wrong shape: {}; expected {}".format(
                            value.shape, p.shape
                        )
                    )
                layer_state += (value,)
            state.append(layer_state)
        self._params = state

    def save(self, filename):
        with open(filename, "wb") as output:
            dump(self.state_dict(), output)

    def load(self, filename):
        with open(filename, "rb") as input:
            self.load_state_dict(load(input))



from .._C_netket.operator import LocalOperator as _LocalOperator

import numpy as _np


def destroy(hilbert, site):
    """
    Builds the boson destruction operator acting on the `site`-th of the
     Hilbert space `hilbert`.

    If `hilbert` is a non-Bosonic space of local dimension M, it is considered
    as a bosonic space of local dimension M.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size

    D = _np.array([_np.sqrt(m) for m in _np.arange(1, N)])
    mat = _np.diag(D, 1)
    return _LocalOperator(hilbert, mat, [site])


def create(hilbert, site):
    """
    Builds the boson creation operator acting on the `site`-th of the
     Hilbert space `hilbert`.

    If `hilbert` is a non-Bosonic space of local dimension M, it is considered
    as a bosonic space of local dimension M.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size

    D = _np.array([_np.sqrt(m) for m in _np.arange(1, N)])
    mat = _np.diag(D, -1)
    return _LocalOperator(hilbert, mat, [site])


def number(hilbert, site):
    """
    Builds the number operator acting on the `site`-th of the
    Hilbert space `hilbert`.

    If `hilbert` is a non-Bosonic space of local dimension M, it is considered
    as a bosonic space of local dimension M.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size

    D = _np.array([m for m in _np.arange(0, N)])
    mat = _np.diag(D, 0)
    return _LocalOperator(hilbert, mat, [site])



from ._C_netket.stats import *


def subtract_mean(x, axis=None, dtype=None, mean_out=None):
    """
    Subtracts the mean of the input array over all but the last dimension
    and over all MPI processes from each entry.

    Args:
        axis: Axis or axes along which the means are computed. The default is to
              compute the mean of the flattened array.
        dtype: Type to use in computing the mean
        mean_out: pre-allocated array to store the mean
    """
    x_mean = mean(x, axis=axis, dtype=dtype, out=mean_out)

    x -= x_mean

    return x


from mpi4py import MPI
import numpy as _np

_MPI_comm = MPI.COMM_WORLD

_n_nodes = _MPI_comm.Get_size()


def mean(a, axis=None, dtype=None, out=None):
    """
    Compute the arithmetic mean along the specified axis and over MPI processes.

    Returns the average of the array elements. The average is taken over the flattened array by default,
    otherwise over the specified axis. float64 intermediate and return values are used for integer inputs.
    """

    out = _np.mean(a, axis=axis, dtype=None, out=out)

    _MPI_comm.Allreduce(MPI.IN_PLACE, out.reshape(-1), op=MPI.SUM)

    out /= float(_n_nodes)

    return out



import sys

import numpy as _np

import netket as _nk
from netket._core import deprecated
from .operator import local_values as _local_values
from netket.stats import (
    statistics as _statistics,
    mean as _mean,
)

from netket.vmc_common import info
from netket.abstract_variational_driver import AbstractVariationalDriver


class Vmc(AbstractVariationalDriver):
    """
    Energy minimization using Variational Monte Carlo (VMC).
    """

    def __init__(
        self, hamiltonian, sampler, optimizer, n_samples, n_discard=None, sr=None
    ):
        """
        Initializes the driver class.

        Args:
            hamiltonian: The Hamiltonian of the system.
            sampler: The Monte Carlo sampler.
            optimizer: Determines how optimization steps are performed given the
                bare energy gradient. This parameter supports three different kinds of inputs,
                which are described in the docs of `make_optimizer_fn`.
            n_samples: Number of Markov Chain Monte Carlo sweeps to be
                performed at each step of the optimization.
            n_discard (int, optional): Number of sweeps to be discarded at the
                beginning of the sampling, at each step of the optimization.
                Defaults to 10% of the number of samples allocated to each MPI node.
            sr (SR, optional): Determines whether and how stochastic reconfiguration
                is applied to the bare energy gradient before performing applying
                the optimizer. If this parameter is not passed or None, SR is not used.

        Example:
            Optimizing a 1D wavefunction with Variational Monte Carlo.

            >>> import netket as nk
            >>> SEED = 3141592
            >>> g = nk.graph.Hypercube(length=8, n_dim=1)
            >>> hi = nk.hilbert.Spin(s=0.5, graph=g)
            >>> ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
            >>> ma.init_random_parameters(seed=SEED, sigma=0.01)
            >>> ha = nk.operator.Ising(hi, h=1.0)
            >>> sa = nk.sampler.MetropolisLocal(machine=ma)
            >>> op = nk.optimizer.Sgd(learning_rate=0.1)
            >>> vmc = nk.Vmc(ha, sa, op, 200)

        """
        super(Vmc, self).__init__(
            sampler.machine, optimizer, minimized_quantity_name="Energy"
        )

        self._ham = hamiltonian
        self._sampler = sampler
        self._sr = sr
        if sr is not None:
            self._sr.is_holomorphic = sampler.machine.is_holomorphic

        self._npar = self._machine.n_par

        self._batch_size = sampler.sample_shape[0]

        self.n_samples = n_samples
        self.n_discard = n_discard

    @property
    def n_samples(self):
        return self._n_samples

    @n_samples.setter
    def n_samples(self, n_samples):
        if n_samples <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples={}".format(n_samples)
            )
        self._n_samples = n_samples
        n_samples_chain = int(_np.ceil((n_samples / self._batch_size)))
        self._n_samples_node = int(_np.ceil(n_samples_chain / _nk.MPI.size()))

        self._samples = _np.ndarray(
            (self._n_samples_node, self._batch_size, self._ham.hilbert.size)
        )

        self._der_logs = _np.ndarray(
            (self._n_samples_node, self._batch_size, self._npar), dtype=_np.complex128
        )

        self._grads = _np.empty(
            (self._n_samples_node, self._machine.n_par), dtype=_np.complex128
        )

    @property
    def n_discard(self):
        return self._n_discard

    @n_discard.setter
    def n_discard(self, n_discard):
        if n_discard is not None and n_discard < 0:
            raise ValueError(
                "Invalid number of discarded samples: n_discard={}".format(n_discard)
            )
        self._n_discard = (
            n_discard
            if n_discard != None
            else self._n_samples_node * self._batch_size // 10
        )

    def _forward_and_backward(self):
        """
        Performs a number of VMC optimization steps.

        Args:
            n_steps (int): Number of steps to perform.
        """

        self._sampler.reset()

        # Burnout phase
        for _ in self._sampler.samples(self._n_discard):
            pass

        # Generate samples and store them
        for i, sample in enumerate(self._sampler.samples(self._n_samples_node)):
            self._samples[i] = sample

        # Compute the local energy estimator and average Energy
        eloc, self._loss_stats = self._get_mc_stats(self._ham)

        # Perform update
        if self._sr:
            # When using the SR (Natural gradient) we need to have the full jacobian
            # Computes the jacobian
            for i, sample in enumerate(self._samples):
                self._der_logs[i] = self._machine.der_log(sample)

            # flatten MC chain dimensions:
            self._der_logs = self._der_logs.reshape(-1, self._npar)

            # Center the local energy
            eloc -= _mean(eloc)

            # Center the log derivatives
            self._der_logs -= _mean(self._der_logs, axis=0)

            # Compute the gradient
            self._grads = _np.conjugate(self._der_logs) * eloc.reshape(-1, 1)

            grad = _mean(self._grads, axis=0)

            dp = _np.empty(self._npar, dtype=_np.complex128)

            self._sr.compute_update(self._der_logs, grad, dp)

            self._der_logs = self._der_logs.reshape(
                self._n_samples_node, self._batch_size, self._npar
            )
        else:
            # Computing updates using the simple gradient
            # Center the local energy
            eloc -= _mean(eloc)

            for x, eloc_x, grad_x in zip(self._samples, eloc, self._grads):
                self._machine.vector_jacobian_prod(x, eloc_x, grad_x)

            grad = _mean(self._grads, axis=0) / float(self._batch_size)
            dp = grad

        return dp

    @property
    def energy(self):
        """
        Return MCMC statistics for the expectation value of observables in the
        current state of the driver.
        """
        return self._loss_stats

    def _estimate_stats(self, obs):
        return self._get_mc_stats(obs)[1]

    def reset(self):
        self._sampler.reset()
        super().reset()

    def _get_mc_stats(self, op):
        loc = _np.empty(self._samples.shape[0:2], dtype=_np.complex128)
        for i, sample in enumerate(self._samples):
            _local_values(op, self._machine, sample, out=loc[i])

        return loc, _statistics(loc)

    def __repr__(self):
        return "Vmc(step_count={}, n_samples={}, n_discard={})".format(
            self.step_count, self.n_samples, self.n_discard
        )

    def info(self, depth=0):
        lines = [
            "{}: {}".format(name, info(obj, depth=depth + 1))
            for name, obj in [
                ("Hamiltonian", self._ham),
                ("Machine", self._machine),
                ("Optimizer", self._optimizer_desc),
                ("SR solver", self._sr),
            ]
        ]
        return "\n  ".join([str(self)] + lines)



import numpy as _np
from netket import random as _random

from .abstract_sampler import AbstractSampler
from .metropolis_hastings import *
from .._C_netket import sampler as c_sampler

from numba import jit, int64, float64
from .._jitclass import jitclass

@jitclass([("clusters", int64[:, :])])
class _exchange_kernel:
    def __init__(self, distances, d_max):
        clusters = []
        size = distances.shape[0]
        for i in range(size):
            for j in range(i + 1, size):
                if(distances[i][j] <= d_max):
                    clusters.append((i, j))

        self.clusters = _np.empty((len(clusters), 2), dtype=int64)

        for i, cluster in enumerate(clusters):
            self.clusters[i] = _np.asarray(cluster)

    def apply(self, state, state_1, log_prob_corr):

        clusters_size = self.clusters.shape[0]

        for k in range(state.shape[0]):
            state_1[k] = state[k]

            # pick a random cluster
            cl = _random.randint(0, clusters_size)

            # sites to be exchanged
            si = self.clusters[cl][0]
            sj = self.clusters[cl][1]

            state_1[k, si], state_1[k, sj] = state[k, sj], state[k, si]

        log_prob_corr[:] = 0.0


class MetropolisExchange(AbstractSampler):
    """
    This sampler acts locally only on two local degree of freedom :math:`s_i` and :math:`s_j`,
    and proposes a new state: :math:`s_1 \dots s^\prime_i \dots s^\prime_j \dots s_N`,
    where in general :math:`s^\prime_i \neq s_i` and :math:`s^\prime_j \neq s_j`.
    The sites :math:`i` and :math:`j` are also chosen to be within a maximum graph
    distance of :math:`d_{\mathrm{max}}`.

    The transition probability associated to this sampler can
    be decomposed into two steps:

    1. A pair of indices :math:`i,j = 1\dots N`, and such
       that :math:`\mathrm{dist}(i,j) \leq d_{\mathrm{max}}`,
       is chosen with uniform probability.
    2. The sites are exchanged, i.e. :math:`s^\prime_i = s_j` and :math:`s^\prime_j = s_i`.

    Notice that this sampling method generates random permutations of the quantum
    numbers, thus global quantities such as the sum of the local quantum numbers
    are conserved during the sampling.
    This scheme should be used then only when sampling in a
    region where :math:`\sum_i s_i = \mathrm{constant}` is needed,
    otherwise the sampling would be strongly not ergodic.
    """

    def __init__(self, machine, d_max=1, n_chains=16, sweep_size=None, batch_size=None):
        """
        Args:
              machine: A machine :math:`\Psi(s)` used for the sampling.
                       The probability distribution being sampled
                       from is :math:`F(\Psi(s))`, where the function
                       :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.

              d_max: The maximum graph distance allowed for exchanges.
              n_chains: The number of Markov Chain to be run in parallel on a single process.
              sweep_size: The number of exchanges that compose a single sweep.
                          If None, sweep_size is equal to the number of degrees of freedom (n_visible).
              batch_size: The batch size to be used when calling log_val on the given Machine.
                          If None, batch_size is equal to the number Markov chains (n_chains).


        Examples:
              Sampling from a RBM machine in a 1D lattice of spin 1/2, using
              nearest-neighbours exchanges.

              >>> import netket as nk
              >>>
              >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
              >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
              >>>
              >>> # RBM Spin Machine
              >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
              >>>
              >>> # Construct a MetropolisExchange Sampler
              >>> sa = nk.sampler.MetropolisExchange(machine=ma)
              >>> print(sa.machine.hilbert.size)
              100
        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.MetropolisExchange(
                machine=machine,
                n_chains=n_chains,
                d_max=d_max,
                sweep_size=sweep_size,
                batch_size=batch_size,
            )
        else:
            self.sampler = PyMetropolisHastings(
                machine,
                _exchange_kernel(
                    _np.asarray(machine.hilbert.graph.distances), d_max
                ),
                n_chains,
                sweep_size,
                batch_size,
            )
        super().__init__(machine, n_chains)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance


class MetropolisExchangePt(AbstractSampler):
    """
    This sampler performs parallel-tempering
    moves in addition to the local moves implemented in `MetropolisExchange`.
    The number of replicas can be chosen by the user.
    """

    def __init__(
        self, machine, d_max=1, n_replicas=16, sweep_size=None, batch_size=None
    ):
        """
        Args:
            machine: A machine :math:`\Psi(s)` used for the sampling.
                     The probability distribution being sampled
                     from is :math:`F(\Psi(s))`, where the function
                     :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
            d_max: The maximum graph distance allowed for exchanges.
            n_replicas: The number of replicas used for parallel tempering.
            sweep_size: The number of exchanges that compose a single sweep.
                        If None, sweep_size is equal to the number of degrees of freedom (n_visible).

        Examples:
            Sampling from a RBM machine in a 1D lattice of spin 1/2, using
            nearest-neighbours exchanges.

            >>> import netket as nk
            >>>
            >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
            >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
            >>>
            >>> # RBM Spin Machine
            >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
            >>>
            >>> # Construct a MetropolisExchange Sampler with parallel tempering
            >>> sa = nk.sampler.MetropolisExchangePt(machine=ma,n_replicas=24)
            >>> print(sa.machine.hilbert.size)
            100
        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.MetropolisExchangePt(
                machine=machine,
                n_replicas=n_replicas,
                d_max=d_max,
                sweep_size=sweep_size,
            )
        else:
            raise ValueError(
                """Parallel Tempering samplers are not yet implemented
                for pure python machines"""
            )
        super().__init__(machine, 1)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance



import abc
import numpy as _np


class AbstractSampler(abc.ABC):
    """Abstract class for NetKet samplers"""

    def __init__(self, machine, sample_size=None):
        super().__init__()
        self.machine = machine

        self.sample_size = sample_size if sample_size != None else 1

        self.sample_shape = (sample_size, machine.hilbert.size)

        self.reset(True)

    def __iter__(self):
        return self

    @abc.abstractmethod
    def __next__(self):
        pass

    @abc.abstractmethod
    def reset(self, init_random=False):
        pass

    @property
    def machine_pow(self):
        return 2.0

    @machine_pow.setter
    def machine_pow(self, m_power):
        raise NotImplementedError

    def samples(self, n_max, init_random=False):

        self.reset(init_random)

        n = 0
        while n < n_max:
            yield self.__next__()
            n += 1

    def generate_samples(self, n_samples, init_random=False, samples=None):
        self.reset(init_random)

        if samples is None:
            samples = _np.zeros(
                (n_samples, self.sample_shape[0], self.sample_shape[1]))

        for k, sample in enumerate(self.samples(n_samples)):
            samples[k] = sample
        return samples



import sys

import numpy as _np

import netket as _nk
from netket._core import deprecated
from .operator import local_values as _local_values
from .operator import _rotated_grad_kernel
from ._C_netket.utils import random_engine, rand_uniform_int

from netket.stats import (
    statistics as _statistics,
    covariance_sv as _covariance_sv,
    subtract_mean as _subtract_mean,
    mean as _mean,
)

from netket.vmc_common import info
from netket.abstract_variational_driver import AbstractVariationalDriver


class Qsr(AbstractVariationalDriver):
    """
    Quantum State Reconstruction for pure states.
    This implements the algorithm introduced in
    Torlai,et al. Nature Phys 14, 447–450 (2018).
    """

    def __init__(
        self,
        sampler,
        optimizer,
        samples,
        rotations,
        bases,
        n_samples,
        n_samples_data,
        n_discard=None,
        sr=None,
    ):
        """
        Initializes the driver class.

        Args:
            sampler: The Monte Carlo sampler.
            optimizer: Determines how optimization steps are performed given the
                bare energy gradient. This parameter supports three different kinds of inputs,
                which are described in the docs of `make_optimizer_fn`.
            samples: An array of training samples from which the wave function is to be reconstructed.
                Shape is (n_training_samples,hilbert.size).
            rotations: A list of `netket.Operator` defining the unitary rotations defining the basis in which
                the samples are given.
            bases: An array of integers of shape (n_training_samples) containing the index of the corresponding rotation.
                If bases[i]=k, for example, then the sample in samples[i] is measured in the basis defined by rotations[k].
            n_samples (int): Number of sampling sweeps to be
                performed at each step of the optimization when sampling from the model wave-function.
            n_samples_data (int): Number of sampling steps to be
                performed at each step of the optimization when sampling from the given data.
            n_discard (int, optional): Number of sweeps to be discarded at the
                beginning of the sampling, at each step of the optimization.
                Defaults to 10% of the number of samples allocated to each MPI node.
            sr (SR, optional): Determines whether and how stochastic reconfiguration
                is applied to the bare energy gradient before performing applying
                the optimizer. If this parameter is not passed or None, SR is not used.

        """
        super(Qsr, self).__init__(sampler.machine, optimizer)

        self._sampler = sampler
        self._sr = sr
        if sr is not None:
            self._sr.is_holomorphic = sampler.machine.is_holomorphic

        self._rotations = rotations
        self._t_samples = _np.asarray(samples)
        self._bases = _np.asarray(bases)

        self._npar = self._machine.n_par

        self._batch_size = sampler.sample_shape[0]
        self._hilbert = self._machine.hilbert

        self.n_samples = n_samples
        self.n_discard = n_discard

        self.n_samples_data = n_samples_data

        assert self._t_samples.ndim == 2
        for samp in self._t_samples:
            assert samp.shape[0] == self._hilbert.size

        self._n_training_samples = self._t_samples.shape[0]

        assert self._bases.ndim == 1
        assert self._bases.size == self._n_training_samples

    @property
    def n_samples(self):
        return self._n_samples

    @n_samples.setter
    def n_samples(self, n_samples):
        if n_samples <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples={}".format(n_samples)
            )
        self._n_samples = n_samples
        n_samples_chain = int(_np.ceil((n_samples / self._batch_size)))
        self._n_samples_node = int(_np.ceil(n_samples_chain / _nk.MPI.size()))

        self._samples = _np.ndarray(
            (self._n_samples_node, self._batch_size, self._hilbert.size)
        )

        self._der_logs = _np.ndarray(
            (self._n_samples_node, self._batch_size, self._npar), dtype=_np.complex128
        )

        self._grads = _np.empty(
            (self._n_samples_node, self._machine.n_par), dtype=_np.complex128
        )

    @property
    def n_samples_data(self):
        return self._n_samples_data

    @n_samples_data.setter
    def n_samples_data(self, n_samples_data):
        if n_samples_data <= 0:
            raise ValueError(
                "Invalid number of samples: n_samples_data={}".format(n_samples)
            )
        self._n_samples_data = n_samples_data
        self._n_samples_data_node = int(_np.ceil(n_samples_data / _nk.MPI.size()))

        self._data_grads = _np.empty(
            (self._n_samples_data_node, self._machine.n_par), dtype=_np.complex128
        )

    @property
    def n_discard(self):
        return self._n_discard

    @n_discard.setter
    def n_discard(self, n_discard):
        if n_discard is not None and n_discard < 0:
            raise ValueError(
                "Invalid number of discarded samples: n_discard={}".format(n_discard)
            )
        self._n_discard = (
            n_discard
            if n_discard != None
            else self._n_samples_node * self._batch_size // 10
        )

    def _forward_and_backward(self):
        """
        Perform one or several iteration steps of the Qsr calculation. In each step,
        the gradient will be estimated via negative and positive phase and subsequently,
        the variational parameters will be updated according to the configured method.

        Args:
            n_steps (int): Number of steps to perform.
        """

        # Generate samples from the model
        self._sampler.reset()

        # Burnout phase
        for _ in self._sampler.samples(self._n_discard):
            pass

        # Generate samples and store them
        for i, sample in enumerate(self._sampler.samples(self._n_samples_node)):
            self._samples[i] = sample

        # Randomly select a batch of training data
        rand_ind = _np.empty(self._n_samples_data_node, dtype=_np.intc)

        rand_uniform_int(0, (self._n_training_samples - 1), rand_ind)

        self._data_samples = self._t_samples[rand_ind]
        self._data_bases = self._bases[rand_ind]

        # Perform update
        if self._sr:
            # When using the SR (Natural gradient) we need to have the full jacobian
            # Computes the jacobian
            for i, sample in enumerate(self._samples):
                self._der_logs[i] = self._machine.der_log(sample)

            grad_neg = _mean(self._der_logs.reshape(-1, self._npar), axis=0).conjugate()

            # Positive phase driven by the data
            for x, b_x, grad_x in zip(
                self._data_samples, self._data_bases, self._data_grads
            ):
                self._compute_rotated_grad(x, b_x, grad_x)

            grad_pos = _mean(self._data_grads, axis=0)

            grad = 2.0 * (grad_neg - grad_pos)

            dp = _np.empty(self._npar, dtype=_np.complex128)

            self._sr.compute_update(self._der_logs.reshape(-1, self._npar), grad, dp)
        else:
            # Computing updates using the simple gradient

            # Negative phase driven by the model
            vec_ones = _np.ones(self._batch_size, dtype=_np.complex128) / float(
                self._batch_size
            )
            for x, grad_x in zip(self._samples, self._grads):
                self._machine.vector_jacobian_prod(x, vec_ones, grad_x)

            grad_neg = _mean(self._grads, axis=0)

            # Positive phase driven by the data
            for x, b_x, grad_x in zip(
                self._data_samples, self._data_bases, self._data_grads
            ):
                self._compute_rotated_grad(x, b_x, grad_x)

            grad_pos = _mean(self._data_grads, axis=0)

            dp = 2.0 * (grad_neg - grad_pos)

        return dp

    def _compute_rotated_grad(self, x, basis, out):
        x_primes, mels = self._rotations[basis].get_conn(x)

        log_val_primes = self._machine.log_val(x_primes)

        vec = _np.empty(mels.size, dtype=_np.complex128)

        _rotated_grad_kernel(log_val_primes, mels, vec)

        self._machine.vector_jacobian_prod(x_primes, vec, out)

    # def _rotated_grad_kernel(self, log_val_primes, mels, vec):
    #     #     max_log_val = log_val_primes.real.max()
    #     #     vec = (mels * _np.exp(log_val_primes - max_log_val)).conjugate()
    #     #     vec /= vec.sum()

    def _estimate_stats(self, obs):
        return self._get_mc_stats(obs)[1]

    def reset(self):
        self._sampler.reset()
        super().reset()

    def _get_mc_stats(self, op):
        loc = _np.empty(self._samples.shape[0:2], dtype=_np.complex128)
        for i, sample in enumerate(self._samples):
            _local_values(op, self._machine, sample, out=loc[i])
        return loc, _statistics(loc)

    def __repr__(self):
        return "Sqr(step_count={}, n_samples={}, n_discard={})".format(
            self.step_count, self.n_samples, self.n_discard
        )

    def info(self, depth=0):
        lines = [
            "{}: {}".format(name, info(obj, depth=depth + 1))
            for name, obj in [
                ("Machine", self._machine),
                ("Optimizer", self._optimizer_desc),
                ("SR solver", self._sr),
            ]
        ]
        return "\n  ".join([str(self)] + lines)

    def nll(self, rotations, samples, bases, log_norm=0):
        """
        Negative log-likelihood, :math:`\langle log(|Psi_b(x)|^2) \rangle`,
        where the average is over the given samples, and :math:`b` denotes
        the given bases associated to the samples.

        Args:
            rotations: Vector of unitary transformation corresponding to basis rotations.
            samples: Vector of samples.
            bases: Which bases the samples correspond to.
            log_norm: This should be :math:`log \sum_x |\Psi(x)|^2`. Notice that
                      if the probability disitribution is not normalized,
                      (i.e. log_norm :math:`\neq 0`), a user-supplied log_norm must be
                      provided, otherwise there is no guarantuee that the
                      negative log-likelihood computed here is a meaningful
                      quantity.
        """
        nll = 0.0
        for x, basis in zip(samples, bases):
            x_primes, mels = rotations[basis].get_conn(x)

            log_val_primes = self._machine.log_val(x_primes)

            max_log_val = log_val_primes.real.max()
            psi_rotated = (mels * _np.exp(log_val_primes - max_log_val)).sum()

            nll -= _np.log(_np.square(_np.absolute(psi_rotated))) + 2.0 * max_log_val

        nll /= float(len(samples))

        return _np.mean(_np.atleast_1d(nll)) + log_norm



from .._C_netket.operator import GraphOperator as _GraphOperator

import numpy as _np

def Ising(hilbert, h, J=1.0):
    """
    Constructs a new ``Ising`` given a hilbert space, a transverse field,
    and (if specified) a coupling constant.

    Args:
        hilbert: Hilbert space the operator acts on.
        h: The strength of the transverse field.
        J: The strength of the coupling. Default is 1.0.

    Examples:
        Constructs an ``Ising`` operator for a 1D system.

        >>> import netket as nk
        >>> g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
        >>> hi = nk.hilbert.Spin(s=0.5, graph=g)
        >>> op = nk.operator.Ising(h=1.321, hilbert=hi, J=0.5)
        >>> print(op.hilbert.size)
        20
    """
    sigma_x = _np.array([[0, 1], [1, 0]])
    sz_sz = _np.array([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
    return _GraphOperator(hilbert, siteops=[-h * sigma_x], bondops=[J * sz_sz])


def Heisenberg(hilbert, J=1, sign_rule=None):
    """
    Constructs a new ``Heisenberg`` given a hilbert space.

    Args:
        hilbert: Hilbert space the operator acts on.
        J: The strength of the coupling. Default is 1.
        sign_rule: If enabled, Marshal's sign rule will be used. On a bipartite
                   lattice, this corresponds to a basis change flipping the Sz direction
                   at every odd site of the lattice. For non-bipartite lattices, the
                   sign rule cannot be applied. Defaults to True if the lattice is
                   bipartite, False otherwise.

    Examples:
     Constructs a ``Heisenberg`` operator for a 1D system.

        >>> import netket as nk
        >>> g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
        >>> hi = nk.hilbert.Spin(s=0.5, total_sz=0, graph=g)
        >>> op = nk.operator.Heisenberg(hilbert=hi)
        >>> print(op.hilbert.size)
        20
    """
    if sign_rule is None:
        sign_rule = hilbert.graph.is_bipartite

    sz_sz = _np.array([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
    exchange = _np.array([[0, 0, 0, 0], [0, 0, 2, 0], [0, 2, 0, 0], [0, 0, 0, 0]])
    if sign_rule:
        if not hilbert.graph.is_bipartite:
            raise ValueError("sign_rule=True specified for a non-bipartite lattice")
        heis_term = sz_sz - exchange
    else:
        heis_term = sz_sz + exchange
    return _GraphOperator(hilbert, bondops=[J * heis_term])



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_machine import AbstractMachine
import numpy as _np
from netket.utils import sum_log_cosh_complex
from .._C_netket.machine import RbmSpinKernel

__all__ = ["PyRbm"]


class PyRbm(AbstractMachine):
    """
    __Do not use me in production code!__

    A proof of concept implementation of a complex-valued RBM in pure Python.

    This class can be used as a drop-in replacement for `RbmSpin`.
    """

    def __init__(
        self, hilbert, alpha=None, use_visible_bias=True, use_hidden_bias=True
    ):
        r"""Constructs a new RBM.

        Args:
            hilbert: Hilbert space.
            alpha: `alpha * hilbert.size` is the number of hidden spins.
            use_visible_bias: specifies whether to use a bias for visible
                              spins.
            use_hidden_bias: specifies whether to use a bias for hidden spins.
        """

        n = hilbert.size
        if alpha < 0:
            raise ValueError("`alpha` should be non-negative")
        m = int(round(alpha * n))
        self._w = _np.empty([m, n], dtype=_np.complex128)
        self._a = _np.empty(n, dtype=_np.complex128) if use_visible_bias else None
        self._b = _np.empty(m, dtype=_np.complex128) if use_hidden_bias else None

        self.n_hidden = m
        self.n_visible = n

        self._r = _np.empty((1, m), dtype=_np.complex128, order="C")

        self._npar = (
            self._w.size
            + (self._a.size if self._a is not None else 0)
            + (self._b.size if self._b is not None else 0)
        )

        self._kernel = RbmSpinKernel()

        super().__init__(hilbert)

    @property
    def n_par(self):
        r"""Returns the number of parameters in the machine. We just sum the
        sizes of all the tensors we hold.
        """
        return self._npar

    def log_val(self, x, out=None):
        r"""Computes the logarithm of the wave function given a spin
        configuration ``x``.
        """
        if out is None:
            out = _np.empty(x.shape[0], dtype=_np.complex128)

        self._kernel.log_val(x, out, self._w, self._a, self._b)

        # self._r = x.dot(self._w.T)
        # if self._b is not None:
        #     self._r += self._b
        #
        # sum_log_cosh_complex(self._r, out)
        #
        # if self._a is not None:
        #     out += _np.dot(x, self._a)

        return out

    def der_log(self, x, out=None):

        if out is None:
            out = _np.empty((x.shape[0], self.n_par), dtype=_np.complex128)

        batch_size = x.shape[0]

        i = 0
        if self._a is not None:
            out[:, i : i + x.shape[1]] = x
            i += self.n_visible

        r = _np.dot(x, self._w.T)
        if self._b is not None:
            r += self._b
        _np.tanh(r, out=r)

        if self._b is not None:
            out[:, i : i + self.n_hidden] = r
            i += self.n_hidden

        t = out[:, i : i + self._w.size]
        t.shape = (batch_size, self._w.shape[0], self._w.shape[1])
        _np.einsum("ij,il->ijl", r, x, out=t)

        return out

    def vector_jacobian_prod(self, x, vec, out=None):
        return _np.dot(_np.asmatrix(self.der_log(x)).H, vec, out)

    @property
    def is_holomorphic(self):
        r"""Complex valued RBM a holomorphic function.
        """
        return True

    @property
    def state_dict(self):
        from collections import OrderedDict

        od = OrderedDict()
        if self._a is not None:
            od["a"] = self._a.view()

        if self._b is not None:
            od["b"] = self._b.view()

        od["w"] = self._w.view()

        return od



from __future__ import absolute_import
from .._C_netket.machine import *
from .cxx_machine import *
from .abstract_machine import AbstractMachine

from .py_rbm import *


def _has_jax():
    try:
        import os

        os.environ["JAX_ENABLE_X64"] = "1"
        import jax

        return True
    except ImportError:
        return False


def _has_torch():
    try:
        import torch
        return True
    except ImportError:
        return False


if _has_jax():
    from .jax import *

if _has_torch():
    from .torch import Torch, TorchLogCosh, TorchView


def MPSPeriodicDiagonal(hilbert, bond_dim, symperiod=-1):
    return MPSPeriodic(hilbert, bond_dim, diag=True, symperiod=symperiod)



from .._C_netket.operator import LocalOperator as _LocalOperator

import numpy as _np


def sigmax(hilbert, site):
    """
    Builds the sigma_x operator acting on the `site`-th of the Hilbert
    space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size
    S = (N - 1) / 2

    D = [_np.sqrt((S + 1) * 2 * a - a * (a + 1)) for a in _np.arange(1, N)]
    mat = _np.diag(D, 1) + _np.diag(D, -1)
    return _LocalOperator(hilbert, mat, [site])


def sigmay(hilbert, site):
    """
    Builds the sigma_y operator acting on the `site`-th of the Hilbert
    space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size
    S = (N - 1) / 2

    D = _np.array(
        [1j * _np.sqrt((S + 1) * 2 * a - a * (a + 1)) for a in _np.arange(1, N)]
    )
    mat = _np.diag(D, -1) + _np.diag(-D, 1)
    return _LocalOperator(hilbert, mat, [site])


def sigmaz(hilbert, site):
    """
    Builds the sigma_z operator acting on the `site`-th of the Hilbert
    space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size
    S = (N - 1) / 2

    D = _np.array([2 * m for m in _np.arange(S, -(S + 1), -1)])
    mat = _np.diag(D, 0)
    return _LocalOperator(hilbert, mat, [site])


def sigmam(hilbert, site):
    """
    Builds the $sigma_- = sigma_x - im * sigma_y$ operator acting on the
    `site`-th of the Hilbert space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size
    S = (N - 1) / 2

    S2 = (S + 1) * S
    D = _np.array([_np.sqrt(S2 - m * (m - 1)) for m in _np.arange(S, -S, -1)])
    mat = _np.diag(D, -1)
    return _LocalOperator(hilbert, mat, [site])


def sigmap(hilbert, site):
    """
    Builds the $sigma_- = sigma_x + im * sigma_y$ operator acting on the
    `site`-th of the Hilbert space `hilbert`.

    If `hilbert` is a non-Spin space of local dimension M, it is considered
    as a (M-1)/2 - spin space.

    :param hilbert: The hilbert space
    :param site: the site on which this operator acts
    :return: a nk.operator.LocalOperator
    """
    N = hilbert.local_size
    S = (N - 1) / 2

    S2 = (S + 1) * S
    D = _np.array([_np.sqrt(S2 - m * (m + 1)) for m in _np.arange(S - 1, -(S + 1), -1)])
    mat = _np.diag(D, 1)
    return _LocalOperator(hilbert, mat, [site])



import numpy as _np
from netket import random as _random

from .abstract_sampler import AbstractSampler
from .metropolis_hastings import *
from .._C_netket import sampler as c_sampler


class CustomSampler(AbstractSampler):
    """
    Custom Sampler, where transition operators are specified by the user.
    For the moment, this functionality is limited to transition operators which
    are sums of :math:`k`-local operators:

    .. math::
       \mathcal{M}= \sum_i M_i


    where the move operators :math:`M_i` act on an (arbitrary) subset of sites.

    The operators :math:`M_i` are specified giving their matrix elements, and a list
    of sites on which they act. Each operator :math:`M_i` must be real,
    symmetric, positive definite and stochastic (i.e. sum of each column and line is 1).

    The transition probability associated to a custom sampler can be decomposed into two steps:

    1. One of the move operators :math:`M_i` is chosen with a weight given by the
      user (or uniform probability by default). If the weights are provided,
      they do not need to sum to unity.

    2. Starting from state
      :math:`|n \rangle`, the probability to transition to state
      :math:`|m\rangle` is given by
      :math:`\langle n|  M_i | m \rangle`.
    """

    def __init__(
        self,
        machine,
        move_operators,
        move_weights=None,
        n_chains=16,
        sweep_size=None,
        batch_size=None,
    ):
        """
        Args:
           machine: A machine :math:`\Psi(s)` used for the sampling.
                  The probability distribution being sampled
                  from is :math:`F(\Psi(s))`, where the function
                  :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
           move_operators: The stochastic `LocalOperator`
                :math:`\mathcal{M}= \sum_i M_i` used for transitions.
           move_weights: For each :math:`i`, the probability to pick one of
                the move operators (must sum to one).
           n_chains: The number of Markov Chains to be run in parallel on a single process.
           sweep_size: The number of exchanges that compose a single sweep.
                       If None, sweep_size is equal to the number of degrees of freedom (n_visible).

       Examples:
           Sampling from a RBM machine in a 1D lattice of spin 1/2

           >>> import netket as nk
           >>>
           >>> g=nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
           >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
           >>>
           >>> # RBM Spin Machine
           >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
           >>>
           >>> # Construct a Custom Sampler
           >>> # Using random local spin flips (Pauli X operator)
           >>> X = [[0, 1],[1, 0]]
           >>> move_op = nk.operator.LocalOperator(hilbert=hi,operators=[X] * g.n_sites,acting_on=[[i] for i in range(g.n_sites)])
           >>> sa = nk.sampler.CustomSampler(machine=ma, move_operators=move_op)
        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.CustomSampler(
                machine=machine,
                move_operators=move_operators,
                move_weights=move_weights,
                n_chains=n_chains,
                sweep_size=sweep_size,
                batch_size=batch_size,
            )
        else:
            self.sampler = PyMetropolisHastings(
                machine,
                c_sampler.CustomLocalKernel(move_operators, move_weights),
                n_chains,
                sweep_size,
                batch_size,
            )
        super().__init__(machine, n_chains)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance


class CustomSamplerPt(AbstractSampler):
    """
    This sampler performs parallel-tempering
    moves in addition to the local moves implemented in `CustomSampler`.
    The number of replicas can be chosen by the user.
    """

    def __init__(
        self, machine, move_operators, move_weights=None, n_replicas=16, sweep_size=None
    ):
        """
        Args:
          machine: A machine :math:`\Psi(s)` used for the sampling.
                   The probability distribution being sampled
                   from is :math:`F(\Psi(s))`, where the function
                   :math:`F(X)`, is arbitrary, by default :math:`F(X)=|X|^2`.
          move_operators: The stochastic `LocalOperator`
                   :math:`\mathcal{M}= \sum_i M_i` used for transitions.
          move_weights: For each :math:`i`, the probability to pick one of
                   the move operators (must sum to one).
          n_replicas: The number of replicas used for parallel tempering.
          sweep_size: The number of exchanges that compose a single sweep.
                      If None, sweep_size is equal to the number of degrees of freedom (n_visible).
        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.CustomSamplerPt(
                machine=machine,
                move_operators=move_operators,
                move_weights=move_weights,
                n_replicas=n_replicas,
                sweep_size=sweep_size,
            )
        else:
            raise ValueError(
                """Parallel Tempering samplers are not yet implemented
                            for pure python machines"""
            )
        super().__init__(machine, 1)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_pow):
        self.sampler.machine_pow = m_pow

    @property
    def acceptance(self):
        """The measured acceptance probability."""
        return self.sampler.acceptance



from __future__ import absolute_import
from .._C_netket.operator import *

from . import spin, boson

from .local_values import (
    local_values,
    der_local_values,
)

from .hamiltonian import (
    Ising,
    Heisenberg,
)

from .._C_netket.operator import _rotated_grad_kernel



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file permits loading numba's jitclass without throwing deprecation
# warnings in v > 0.49.

# TODO if numba's minimum version is raised to 0.49, remove this file and
# change all usages to from numba.experimental import jitclass

from pkg_resources import get_distribution

if get_distribution("numba").version < '0.49':
    from numba import jitclass
else:
    from numba.experimental import jitclass



from .._C_netket.operator import (
    _der_local_values_kernel,
    _der_local_values_notcentered_kernel,
    LocalLiouvillian
)


import numpy as _np
from numba import jit


from .._C_netket.machine import DensityMatrix


@jit(nopython=True)
def _local_values_kernel(log_vals, log_val_primes, mels, sections, out):
    low_range = 0
    for i, s in enumerate(sections):
        out[i] = (mels[low_range:s] *
                  _np.exp(log_val_primes[low_range:s] - log_vals[i])).sum()
        low_range = s


def _local_values_impl(op, machine, v, log_vals, out):

    sections = _np.empty(v.shape[0], dtype=_np.int32)
    v_primes, mels = op.get_conn_flattened(v, sections)

    log_val_primes = machine.log_val(v_primes)

    _local_values_kernel(log_vals, log_val_primes, mels, sections, out)


@jit(nopython=True)
def _op_op_unpack_kernel(v, sections, vold):

    low_range = 0
    for i, s in enumerate(sections):
        vold[low_range:s] = v[i]
        low_range = s

    return vold


def _local_values_op_op_impl(op, machine, v, log_vals, out):

    sections = _np.empty(v.shape[0], dtype=_np.int32)
    v_primes, mels = op.get_conn_flattened(v, sections)

    vold = _np.empty((sections[-1], v.shape[1]))
    _op_op_unpack_kernel(v, sections, vold)

    log_val_primes = machine.log_val(v_primes, vold)

    _local_values_kernel(log_vals, log_val_primes, mels, sections, out)


def local_values(op, machine, v, log_vals=None, out=None):
    """
    Computes local values of the operator `op` for all `samples`.

    The local value is defined as
    .. math:: O_{\mathrm{loc}}(x) = \langle x | O | \Psi \rangle / \langle x | \Psi \rangle


            Args:
                op: Hermitian operator.
                v: A numpy array or matrix containing either a single
                    :math:`V = v` or a batch of visible
                    configurations :math:`V = v_1,\dots v_M`.
                    In the latter case, each row of the matrix corresponds to a
                    visible configuration.
                machine: Wavefunction :math:`\Psi`.
                log_vals: A scalar/numpy array containing the value(s) :math:`\Psi(V)`.
                    If not given, it is computed from scratch.
                    Defaults to None.
                out: A scalar or a numpy array of local values of the operator.
                    If not given, it is allocated from scratch and then returned.
                    Defaults to None.

            Returns:
                If samples is given in batches, a numpy array of local values
                of the operator, otherwise a scalar.
    """

    # True when this is the local_value of a densitymatrix times an operator (observable)
    is_op_times_op = isinstance(machine, DensityMatrix) and not isinstance(
        op, LocalLiouvillian
    )

    if log_vals is None:
        if not is_op_times_op:
            log_vals = machine.log_val(v)
        else:
            log_vals = machine.log_val(v, v)

    if not is_op_times_op:
        _impl = _local_values_impl
    else:
        _impl = _local_values_op_op_impl

    if v.ndim == 3:
        assert (
            v.shape[2] == op.hilbert.size
        ), "samples has wrong shape: {}; expected (?, {})".format(
            v.shape, op.hilbert.size
        )

        if out is None:
            out = _np.empty(v.shape[0] * v.shape[1], dtype=_np.complex128)

        _impl(
            op,
            machine,
            v.reshape(-1, op.hilbert.size),
            log_vals.reshape(-1),
            out.reshape(-1),
        )

        return out.reshape(v.shape[0:-1])
    elif v.ndim == 2:
        assert (
            v.shape[1] == op.hilbert.size
        ), "samples has wrong shape: {}; expected (?, {})".format(
            v.shape, op.hilbert.size
        )

        if out is None:
            out = _np.empty(v.shape[0], dtype=_np.complex128)

        _impl(op, machine, v, log_vals, out)

        return out
    elif v.ndim == 1:
        assert v.size == op.hilbert.size, "v has wrong size: {}; expected {}".format(
            v.shape, op.hilbert.size
        )
        if out is None:
            out = _np.empty(1, dtype=_np.complex128)
        else:
            out = _np.atleast_1d(out)

        log_vals = _np.atleast_1d(log_vals)

        _impl(op, machine, v.reshape(1, -1), log_vals.reshape(1, -1), out)
        return out[0]
    raise ValueError(
        "v has wrong dimension: {}; expected either 1, 2 or 3".format(v.ndim)
    )


def _der_local_values_impl(op, machine, v, log_vals, der_log_vals, out):

    vprimes, mels = op.get_conn(v)

    log_val_primes = [machine.log_val(vprime) for vprime in vprimes]

    der_log_primes = [machine.der_log(vprime) for vprime in vprimes]

    _der_local_values_kernel(
        log_vals, log_val_primes, mels, der_log_vals, der_log_primes, out
    )

    # for k, sample in enumerate(v):
    #
    #     lvd = machine.log_val(vprimes[k])
    #
    #     dld = machine.der_log(vprimes[k])
    #
    #     out[k,:] = (((mels[k] * _np.exp(lvd - log_vals[k])) * (dld - der_log_val[k,:])).sum(axis=0)


# TODO: numba or cython to improve performance of this kernel
def der_local_values_notcentered_kernel(log_vals, log_val_p, mels, der_log_p, out):
    for k in range(len(mels)):
        out[k, :] = ((mels[k] * _np.exp(log_val_p[k] - log_vals[k]))[:, _np.newaxis]
                     * der_log_p[k]).sum(axis=0)


def _der_local_values_notcentered_impl(op, machine, v, log_vals, out):

    vprimes, mels = op.get_conn(v)

    log_val_primes = [machine.log_val(vprime) for vprime in vprimes]

    der_log_primes = [machine.der_log(vprime) for vprime in vprimes]

    # Avoid using the C++ kernel because of a pybind11 problem. We are probably
    # copying list-of-numpy arrays when calling C++, which leads to a huge
    # performance degradation
    der_local_values_notcentered_kernel(
        log_vals, log_val_primes, mels, der_log_primes, out
    )


def der_local_values(
    op, machine, v, log_vals=None, der_log_vals=None, out=None, center_derivative=True
):
    """
    Computes the derivative of local values of the operator `op` for all `samples`.

    The local value is defined as
    .. math:: O_{\mathrm{loc}}(x) = \langle x | O | \Psi \rangle / \langle x | \Psi \rangle


            Args:
                op: Hermitian operator.
                v: A numpy array or matrix containing either a single
                    :math:`V = v` or a batch of visible
                    configurations :math:`V = v_1,\dots v_M`.
                    In the latter case, each row of the matrix corresponds to a
                    visible configuration.
                machine: Wavefunction :math:`\Psi`.
                log_vals: A scalar/numpy array containing the value(s) :math:`\Psi(V)`.
                    If not given, it is computed from scratch.
                    Defaults to None.
                der_log_vals: A numpy tensor containing the vector of log-derivative(s) :math:`O_i(V)`.
                    If not given, it is computed from scratch.
                    Defaults to None.
                out: A scalar or a numpy array of local values of the operator.
                    If not given, it is allocated from scratch and then returned.
                    Defaults to None.
                center_derivative: Whever to center the derivatives or not. In the formula above,
                    When this is true/false it is equivalent to setting :math:`\alpha=\{1 / 2\}`.
                    By default `center_derivative=True`, meaning that it returns the correct
                    derivative of the local values. False is mainly used when dealing with liouvillians.

            Returns:
                If samples is given in batches, a numpy ndarray of derivatives of local values
                of the operator, otherwise a 1D array.
    """
    if v.ndim == 3:
        assert (
            v.shape[2] == op.hilbert.size
        ), "samples has wrong shape: {}; expected (?, {})".format(
            v.shape, op.hilbert.size
        )

        if out is None:
            out = _np.empty(
                (v.shape[0] * v.shape[1], machine.n_par), dtype=_np.complex128
            )

        if log_vals is None:
            log_vals = machine.log_val(v)

        if der_log_vals is None and center_derivative is True:
            der_log_vals = machine.der_log(v)

        if center_derivative is True:
            _der_local_values_impl(
                op,
                machine,
                v.reshape(-1, op.hilbert.size),
                log_vals.reshape(-1),
                der_log_vals.reshape(-1, machine.n_par),
                out.reshape(-1, machine.n_par),
            )
        else:
            _der_local_values_notcentered_impl(
                op,
                machine,
                v.reshape(-1, op.hilbert.size),
                log_vals.reshape(-1),
                out.reshape(-1, machine.n_par),
            )

        return out.reshape(v.shape[0], v.shape[1], machine.n_par)
    elif v.ndim == 2:
        assert (
            v.shape[1] == op.hilbert.size
        ), "samples has wrong shape: {}; expected (?, {})".format(
            v.shape, op.hilbert.size
        )

        if out is None:
            out = _np.empty((v.shape[0], machine.n_par), dtype=_np.complex128)

        if log_vals is None:
            log_vals = machine.log_val(v)

        if der_log_vals is None and center_derivative is True:
            der_log_vals = machine.der_log(v)

        if center_derivative is True:
            _der_local_values_impl(op, machine, v, log_vals, der_log_vals, out)
        else:
            _der_local_values_notcentered_impl(op, machine, v, log_vals, out)

        return out
    elif v.ndim == 1:
        assert v.size == op.hilbert.size, "v has wrong size: {}; expected {}".format(
            v.shape, op.hilbert.size
        )
        if out is None:
            out = _np.empty((1, machine.n_par), dtype=_np.complex128)
        else:
            out = _np.atleast_2d(out)

        if log_vals is None:
            log_vals = machine.log_val(v)

        log_vals = _np.atleast_1d(log_vals)

        if der_log_vals is None and center_derivative is True:
            der_log_vals = machine.der_log(v)

        der_log_vals = _np.atleast_2d(der_log_vals)

        if center_derivative is True:
            _der_local_values_impl(
                op,
                machine,
                v.reshape(1, -1),
                log_vals.reshape(1, -1),
                der_log_vals,
                out,
            )
        else:
            _der_local_values_notcentered_impl(
                op, machine, v.reshape(1, -1), log_vals.reshape(1, -1), out
            )

        return out[0, :]
    raise ValueError(
        "v has wrong dimension: {}; expected either 1, 2 or 3".format(v.ndim)
    )



from .._C_netket.machine import Machine
import numpy as _np

class CxxMachine(Machine):
    def __init__(self, hilbert):
        super(CxxMachine, self).__init__(hilbert)

    def _get_parameters(self):
        return _np.concatenate(tuple(p.reshape(-1) for p in self.state_dict().values()))

    def _set_parameters(self, p):
        if p.shape != (self._n_par(),):
            raise ValueError(
                "p has wrong shape: {}; expected ({},)".format(p.shape, self._n_par())
            )
        i = 0
        for x in map(lambda x: x.reshape(-1), self.state_dict().values()):
            _np.copyto(x, p[i : i + x.size])
            i += x.size

    def log_val(self, v, out=None):
        r"""Computes the logarithm of the wave function for a batch of visible
        configurations `v` and stores the result into `out`.

        Args:
            v: Either a
                * vector of `float64` of size `self.n_visible` or
                * a matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination vector of `complex128`. If `v` is a matrix then
                length of `out` should be `v.shape[0]`. If `v` is a vector,
                then length of `out` should be 1.

        Returns:
            A complex number when `v` is a vector and vector when `v` is a
            matrix.


        Subclasses should implement this function.

        Since it is very cumbersome to properly handle all the cases, there is
        a default implementation of `log_val` which uses `_log_val` under the
        hood. `_log_val` has similar semantics to `log_val` except that it is
        guaranteed that `v` is a 2D array and that `out` is a 1D array which is
        never `None`.
        """
        if v.ndim == 2:
            assert (
                v.shape[1] == self.n_visible
            ), "v has wrong shape: {}; expected (?, {})".format(v.shape, self.n_visible)
            if out is None:
                out = _np.empty(v.shape[0], dtype=_np.complex128)
            self._log_val(v, out)
            return out
        elif v.ndim == 1:
            assert v.size == self.n_visible, "v has wrong size: {}; expected {}".format(
                v.shape, self.n_visible
            )
            if out is None:
                out = _np.empty(1, dtype=_np.complex128)
            self._log_val(v.reshape(1, -1), out)
            return out[0]
        raise ValueError(
            "v has wrong dimension: {}; expected either 1 or 2".format(v.ndim)
        )

    def _log_val(self, v, out):
        raise NotImplementedError(
            "override me for the default implementation of log_val to work"
        )

    def der_log(self, v, out=None):
        r"""Computes the gradient of the logarithm of the wavefunction for a
        batch of visible configurations `v` and stores the result into `out`.

        Args:
            v: Either a
                * vector of `float64` of size `self.n_visible` or
                * a matrix of `float64` of shape `(*, self.n_visible)`.
            out: Destination tensor of `complex128`. If `v` is a matrix then of
                `out` should be a matrix of shape `(v.shape[0], self.n_par)`.
                If `v` is a vector, then `out` should be a vector of length
                `self.n_par`.

        Returns:
            `out`

        Subclasses should implement this function.

        Since it is very cumbersome to properly handle all the cases, there is
        a default implementation of `der_log` which uses `_der_log` under the
        hood. `_der_log` has similar semantics to `der_log` except that it is
        guaranteed that both `v` and `out` are 2D arrays (which implies that
        `out` is never `None`).
        """
        if v.ndim == 2:
            assert (
                v.shape[1] == self.n_visible
            ), "v has wrong shape: {}; expected (?, {})".format(v.shape, self.n_visible)
            out_shape = (v.shape[0], self._n_par())
            if out is None:
                out = _np.empty(out_shape, dtype=_np.complex128)
            assert (
                out.shape == out_shape
            ), "out has wrong shape: {}; expected {}".format(out.shape, out_shape)
            self._der_log(v, out)
            return out
        elif v.ndim == 1:
            assert v.size == self.n_visible, "v has wrong size: {}; expected {}".format(
                v.shape, self.n_visible
            )
            if out is None:
                out = _np.empty(self._n_par(), dtype=_np.complex128)
            assert out.shape == (
                self._n_par(),
            ), "out has wrong shape: {}; expected {}".format(
                out.shape, (self._n_par(),)
            )
            self._der_log(v.reshape(1, -1), out.reshape(1, -1))
            return out
        raise ValueError(
            "v has wrong dimension: {}; expected either 1 or 2".format(v.ndim)
        )

    def _der_log(self, v, out):
        raise NotImplementedError

    # def save(self, filename):
    #     pass

    # def load(self, filename):
    #     pass

    def _is_holomorphic(self):
        r"""Returns whether the wave function is holomorphic.
        """
        raise NotImplementedError

    def _n_par(self):
        r"""Returns the number of variational parameters in the machine.

        Subclasses should implement this function, but to actually get the
        number of parameters, use `self.n_par` attribute.
        """
        raise NotImplementedError



import numpy as _np
from numba import jit, objmode
from mpi4py import MPI


@jit
def seed(seed=None):

    with objmode(derived_seed="int64"):
        comm = MPI.COMM_WORLD
        size = comm.Get_size()
        rank = comm.Get_rank()

        if rank == 0:
            _np.random.seed(seed)
            derived_seed = _np.random.randint(0, 1 << 32, size=size)
        else:
            derived_seed = None

        derived_seed = comm.scatter(derived_seed, root=0)

    _np.random.seed(derived_seed)


@jit
def uniform(low=0.0, high=1.0):
    return _np.random.uniform(low, high)


@jit
def randint(low, high):
    return _np.random.randint(low, high)


# By default, the generator is initialized with a random seed (on node 0)
# and then propagated correctly to the other nodes
seed(None)



from ._C_netket.dynamics import *
from . import _core


@_core.deprecated("function has been renamed to `timestepper`")
def create_timestepper(*args, **kwargs):
    return timestepper(*args, **kwargs)



# Copyright 2019 The Simons Foundation, Inc. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import functools
import inspect


def deprecated(reason=None):
    """
    This is a decorator which can be used to mark functions as deprecated. It
    will result in a warning being emitted when the function is used.
    """

    def decorator(func):
        object_type = "class" if inspect.isclass(func) else "function"
        message = "Call to deprecated {} {!r}".format(object_type, func.__name__)
        if reason is not None:
            message += " ({})".format(reason)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(message, category=FutureWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def warn_deprecation(message):
    """
    This is a function that sends a deprecation warning to the user about a
    function that will is now deprecated and will be removed in a future
    major release.

    :param message: A mandatory message documenting the deprecation.
    """
    warnings.warn(message, category=FutureWarning, stacklevel=2)



import numpy as _np
from .abstract_sampler import AbstractSampler

from .._C_netket import sampler as c_sampler
from .._C_netket.utils import random_engine, rand_uniform_real


class PyExactSampler(AbstractSampler):
    """
    Pure Python version of ExactSampler. See ExactSampler for more details.
    """

    def __init__(self, machine, sample_size=16):
        """
         Constructs a new ``PyExactSampler`` given a machine.

         Args:
             machine: A machine $$\Psi(s)$$ used for the sampling.
                      The probability distribution being sampled
                      from is $$F(\Psi(s))$$, where the function
                      $$F(X)$$, is arbitrary, by default $$F(X)=|X|^2$$.

             sample_size: The number of independent samples to be generated at each invocation of __next__.
        """
        self.hilbert = machine.hilbert
        self.machine_pow = 2.0
        super().__init__(machine, sample_size)

    def reset(self, init_random=False):
        self._prob = _np.exp(self.machine_pow * self.machine.to_array().real())
        self._prob /= self._prob.sum()

    def __next__(self):
        numbers = _np.random.choice(
            self._prob.size, size=self.sample_shape[0], replace=True, p=self._prob
        )
        return self.hilbert.number_to_state(numbers)

    @property
    def machine_pow(self):
        return self._machine_pow

    @machine_pow.setter
    def machine_pow(self, m_power):
        self._machine_pow = m_power
        self.reset()


class ExactSampler(AbstractSampler):
    """
    This sampler generates i.i.d. samples from $$|\Psi(s)|^2$$.
    In order to perform exact sampling, $$|\Psi(s)|^2$$ is precomputed an all
    the possible values of the quantum numbers $$s$$. This sampler has thus an
    exponential cost with the number of degrees of freedom, and cannot be used
    for large systems, where Metropolis-based sampling are instead a viable
    option.
    """

    def __init__(self, machine, sample_size=16):
        """
         Constructs a new ``ExactSampler`` given a machine.

         Args:
             machine: A machine $$\Psi(s)$$ used for the sampling.
                      The probability distribution being sampled
                      from is $$F(\Psi(s))$$, where the function
                      $$F(X)$$, is arbitrary, by default $$F(X)=|X|^2$$.

             sample_size: The number of independent samples to be generated at each invocation of __next__.

         Examples:
             Exact sampling from a RBM machine in a 1D lattice of spin 1/2

             ```python
             >>> import netket as nk
             >>>
             >>> g=nk.graph.Hypercube(length=8,n_dim=1,pbc=True)
             >>> hi=nk.hilbert.Spin(s=0.5,graph=g)
             >>>
             >>> # RBM Spin Machine
             >>> ma = nk.machine.RbmSpin(alpha=1, hilbert=hi)
             >>>
             >>> sa = nk.sampler.ExactSampler(machine=ma)

             ```
        """
        if "_C_netket.machine" in str(type(machine)):
            self.sampler = c_sampler.ExactSampler(
                machine=machine, sample_size=sample_size
            )
        else:
            self.sampler = PyExactSampler(
                machine=machine, sample_size=sample_size)
        super().__init__(machine, sample_size)

    def reset(self, init_random=False):
        self.sampler.reset(init_random)

    def __next__(self):
        return self.sampler.__next__()

    @property
    def machine_pow(self):
        return self.sampler.machine_pow

    @machine_pow.setter
    def machine_pow(self, m_power):
        self.sampler.machine_pow = m_power
